# Copyright (c) 2019-2020 Jonathan Sambrook and Codethink Ltd.
#
#    This file is part of Topplot.
#
#    Topplot is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Topplot is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Topplot.  If not, see <https://www.gnu.org/licenses/>.
#
# ------------------------------------------------------------------------------

import re
import sys
import tempfile

# ------------------------------------------------------------------------------
# Exit with error, displaying msg to stderr


def die(msg):
    print(f"ERR: {msg}", file=sys.stderr)
    sys.exit(1)


# ------------------------------------------------------------------------------
# Display labelled warning message to stderr


def warn(msg):
    print(f"WARN: {msg}", file=sys.stderr)


# ------------------------------------------------------------------------------
# Display labelled informational message to stderr


def info(msg):
    print(f"INFO: {msg}", file=sys.stderr)


# ------------------------------------------------------------------------------
# Convert DD:HH:MM:SS to seconds (re based to flexibly handle optional day/hour/minutes)
# Dies on failure to parse!

RE_DHMS_TO_SEC = re.compile(
    r"^((?P<d>\d+):){0,1}?((?P<h>\d+):){0,1}?((?P<m>\d+):){0,1}?(?P<s>\d\d)$"
)


def dhms_to_sec(text):
    m = RE_DHMS_TO_SEC.match(text)

    if not m:
        die(f"'{text}' doesn't parse as a timestamp with format [[[D:]HH:]MM:]SS")

    groups = m.groupdict()
    d = int(groups["d"]) if groups["d"] else 0
    h = int(groups["h"]) if groups["h"] else 0
    m = int(groups["m"]) if groups["m"] else 0
    s = int(groups["s"])
    # print(f"{text} → d: {d}  h: {h}  m: {m}  s: {s}")
    return d * 24 * 3600 + h * 3600 + m * 60 + s


# ------------------------------------------------------------------------------


def is_dir_writable(path):
    try:
        testfile = tempfile.TemporaryFile(dir=path)
        testfile.close()
    except OSError as e:
        e.filename = path
        raise
    return True


# ------------------------------------------------------------------------------
# vi: sw=4:ts=4:et



import datetime as dt
from io import StringIO
import re
import sys

import matplotlib as mpl
import pandas as pd


class CpuFiles:
    def __init__(self, config, directory, poi_combined_data_filename):
        self.config = config
        self.dir = directory
        self.poi_combined_data_filename = poi_combined_data_filename

        self.files = []
        self.poi_combined_data = {}
        self.registry = {}
        for core in range(0, self.config.cores):
            self.files.append(open(f"{directory}/cpu{core}_process.data", "w"))

    def close(self):
        if self.files:
            for f in self.files:
                f.close()
            self.files = []

            poi_keys = []
            for core in range(0, self.config.cores):
                poi_keys.append(f"cpu{core} poi cpu")
                poi_keys.append(f"cpu{core} poi mem")

            for core in range(0, self.config.cores):
                if self.config.plot_poi_cpu_sum or self.config.plot_poi_mem_sum:
                    # Generate data file
                    write_datafile(
                        self.poi_combined_data_filename,
                        self.poi_combined_data,
                        poi_keys,
                    )

    def write(self, core, header, txt):
        if core not in self.registry:
            self.registry[core] = self.files[core]
            if self.config.allow_write:
                self.files[core].write(header)

        if self.config.allow_write:
            self.files[core].write(txt)

    def seal_register(self):
        if self.config.allow_write:
            for cpu_id in self.registry:
                self.registry[cpu_id].write("\n")
        self.registry = {}

    def init_poi_timestamp(self, timestamp):
        self.poi_combined_data[timestamp] = {}
        for i in range(0, self.config.cores):
            self.poi_combined_data[timestamp][f"cpu{i} poi cpu"] = 0
            self.poi_combined_data[timestamp][f"cpu{i} poi mem"] = 0

    def add_poi_cpu(self, core, timestamp, value):
        if timestamp not in self.poi_combined_data:
            self.init_poi_timestamp(timestamp)
        self.poi_combined_data[timestamp][f"cpu{core} poi cpu"] += float(value)

    def add_poi_mem(self, core, timestamp, value):
        if timestamp not in self.poi_combined_data:
            self.init_poi_timestamp(timestamp)
        self.poi_combined_data[timestamp][f"cpu{core} poi mem"] += float(value)


# ------------------------------------------------------------------------------

MIDNIGHT = dt.datetime.combine(dt.date.today(), dt.time())


def from_csv(f):
    df = pd.read_csv(f, sep=" ")
    # Using Pandas datetimes lead to a weird issue with plotting dataframes on
    # multiplot figures.
    # Using MPL datetimes just works
    df.timestamp = df.timestamp.apply(
        lambda x: mpl.dates.date2num((MIDNIGHT + dt.timedelta(seconds=x)))
    )
    df.set_index("timestamp", inplace=True)
    return df


# --------------------------------------------------------------------------


def dataframe_from_process_graph_data(datafile):
    df_mono = None
    lines = []

    def closeoff(lines):
        # close off and process current set
        nonlocal df_mono
        csv = StringIO("".join(lines))
        df = from_csv(csv)
        if df_mono is None:
            df_mono = df
        else:
            df_mono = pd.concat([df_mono, df], axis=1)

    with open(datafile) as poi_data:
        process_name = None
        lines = []
        re_starts_with_timestamp_header = re.compile(r'^timestamp "(.*) - cpu"')
        re_parse_data = re.compile(r"^(\d+) ([0-9.]+) ([0-9.]+)\n$")
        for line in poi_data.readlines():
            if line == "\n":
                continue
            header_match = re_starts_with_timestamp_header.match(line)
            if header_match:
                if process_name is not None:
                    closeoff(lines)

                process_name = header_match.group(1)
                lines = [line]
            else:
                match = re_parse_data.match(line)
                if match:
                    lines.append(line)
                else:
                    print(
                        f"ERR: poi graphing Don't know what to do with line: '{line}'",
                        file=sys.stderr,
                    )

    if lines:
        closeoff(lines)
    else:
        df_mono = pd.DataFrame()

    return df_mono


def write_datafile(filename, source, keys):
    with open(filename, "w") as data_file:
        # Header
        line = '"timestamp"'
        for key in keys:
            key = key.replace("_", " ")
            line = f'{line} "{key}"'
        data_file.write(line + "\n")

        # Data
        # First with an integrated timestamp
        if isinstance(source, list):
            for entry in source:
                added = False
                line = entry["timestamp"]
                for key in keys:
                    added = True
                    line = f"{line} {entry[key]}"
                if added:
                    data_file.write(line + "\n")
        elif isinstance(source, dict):
            # Otherwise the key *is* the timestamp
            for timestamp, columns in source.items():
                line = timestamp
                for key in keys:
                    line = f"{line} {columns[key]}"
                data_file.write(line + "\n")



import os
import re
import sys
import tkinter as tk
from tkinter import Tk

from .gui_utils import center_window_geometry


class ProgressWindow:
    root: Tk

    def __init__(self):
        bg = (  # Equivalent to value of Grapher's colours["fig_face"] = (0.95, 0.95, 0.90).
            "#F2F2E6"  # Keep in sync.
        )

        self.root = tk.Tk()

        png = os.path.join(os.path.dirname(__file__), "logo.png")
        self.logo = tk.PhotoImage(file=png).subsample(5)
        self.root.iconphoto(True, self.logo)

        self.root.attributes("-alpha", 0.0)
        self.root.configure(bg=bg)
        self.root.update_idletasks()  # Get up to date with current situation

        # Flip into fullscreen, stash geometry, flip out
        self.root.attributes("-fullscreen", True)
        self.root.state("iconic")
        self.screen_geometry = self.root.winfo_geometry()
        self.root.attributes("-fullscreen", False)

        # Setup fixed sized, always on top window
        self.root.title("topplot")
        self.root.geometry("500x150")
        self.root.resizable(False, False)
        self.root.attributes("-topmost", 1)
        self.root.state("normal")

        if sys.platform.startswith("win"):
            # Don't add screen furniture
            self.root.overrideredirect(1)  # Messes up centering on my Linux WM

        # Setup frame and contained logo and label
        frame = tk.Frame(self.root, bd=1, relief="raised")
        frame.config(bg=bg)
        frame.place(relx=0.5, rely=0.5, anchor="center")

        logo = tk.Label(frame, image=self.logo)
        logo.config(bg=bg)
        logo.pack(side=tk.TOP, padx=10, pady=10)

        self.status = tk.Label(
            frame, text="topplot is initializing...                 ",
        )
        self.status.config(bg=bg)
        self.status.pack(expand=True, side=tk.LEFT)

        frame.pack(fill=tk.BOTH, expand=True)

        self.root.update_idletasks()
        self.center()

        self.root.attributes("-alpha", 1.0)

    # ----------------------------------------------------------------------

    def center(self, win=None):
        if self.root is not None:
            if win is None:
                win = self.root
                self.root.update()
            else:
                self.root.update_idletasks()

            win_width = win.winfo_width()
            win_height = win.winfo_height()

            screen_width, screen_height = self.get_curr_screen_dimensions()

            # print(f"{win_width=}    {win_height=}")
            # print(f"{screen_width=}    {screen_height=}")

            new_geometry = center_window_geometry(
                win_width,
                win_height,
                screen_width,
                screen_height,
                self.root.winfo_pointerx(),
                self.root.winfo_pointery(),
            )

            # print(f"{new_geometry=}")

            win.geometry(new_geometry)

            win_width = win.winfo_width()
            win_height = win.winfo_height()

    # ----------------------------------------------------------------------

    def hide(self):
        if self.root is not None:
            self.root.withdraw()

    # ----------------------------------------------------------------------

    def show(self):
        if self.root is not None:
            self.root.deiconify()
            self.root.update()

    # ----------------------------------------------------------------------

    def destroy(self):
        if self.root is not None:
            self.root.quit()
            self.root.destroy()
            self.root = None

    # ----------------------------------------------------------------------

    def update_status(self, msg):
        if self.root is not None:
            self.status.configure(text=f"topplot is {msg}...")
            self.root.update()

    # ----------------------------------------------------------------------

    def get_curr_screen_dimensions(self, scale=1.0):
        if self.screen_geometry is None:
            return None

        m = re.match(
            r"^(?P<width>\d+)x(?P<height>\d+)\+(?P<position>\d+\+\d+)",
            self.screen_geometry,
        )
        groups = m.groupdict()
        return (
            int(int(groups["width"]) * scale),
            int(int(groups["height"]) * scale),
        )



import re
import sys
from typing import Dict, List, Tuple

from .re_variants import Re_Variants
from ..toptype import TopType

# ------------------------------------------------------------------------------

MUSTHAVE_COLUMNS: Dict[TopType, List[str]] = {
    TopType.PROCPS: ["%CPU", "COMMAND", "%MEM", "PID"],
    TopType.BUSYBOX: ["%CPU", "COMMAND", "%VSZ", "PID"],
}

# ------------------------------------------------------------------------------


def _setup_regex(self):
    self._re_ignore_processes_regex = None
    if self.config.ignore_processes_regex:
        flags = 0
        if self.config.ignore_case:
            flags = re.IGNORECASE

        self._re_ignore_processes_regex = re.compile(
            self.config.ignore_processes_regex, flags
        )

    # --------------------------------------------------------------------------
    # Different versions of top require slightly different handling
    #
    # This is being handled by setting variables when what parses (!) for version
    # detection occurs.
    #
    # This is performed by optionally passing tuples of:
    #    ((var1_name, var1_value), (varN_name, varN_value))
    # in with regexes to Re_Variants instances.

    # --------------------------------------------------------------------------
    # Precompile regexps
    #
    # The group names, i.e the "word" in (?P<word>pattern), are used later
    # on as dictionary keys

    # top - 06:40:46 up 0 min,  0 users,  load average: 20.84, 5.41, 1.83
    self._re_top_procps = re.compile(
        r"^top - (?P<timestamp>[^ ]+) .*load average: (?P<load_average>[0-9.]+), .*"
    )

    # Did I mention yet how annoying I find BusyBox in so many ways?
    # Replacing common utilities with such hideous simulcra is an abomination :)
    # No timestamp, even in batch output. What were they thinking?
    # Mem: 27436964K used, 5056536K free, 317416K shrd, 1433660K buff, 11810180K cached
    self._re_top_busybox = re.compile(
        r"^Mem:"
        r" (?P<mem_used>[0-9]+)K used,"
        r" (?P<mem_free>[0-9]+)K free,"
        r" (?P<mem_shared>[0-9]+)K shrd,"
        r" (?P<mem_buffers>[0-9]+)K buff,"
        r" (?P<mem_cached>[0-9]+)K cached.*"
    )

    scope = locals()  # pylint: disable=used-before-assignment

    # Tasks: 301 total,  23 running, 209 sleeping,   0 stopped,   0 zombie
    self._re_procps_tasks = Re_Variants(
        "re_procps_tasks",
        re.compile(
            r"^Tasks:"
            r" (?P<task_total>[0-9]+) total,"
            r" +(?P<task_running>[0-9]+) running,"
            r" +(?P<task_sleeping>[0-9]+) sleeping,"
            r" +(?P<task_stopped>[0-9]+) stopped,"
            r" +(?P<task_zombie>[0-9]+) zombie"
        ),
    )

    # Threads: 1529 total,   4 running, 1525 sleeping,   0 stopped,   0 zombie
    self._re_procps_tasks.append(
        re.compile(
            r"^Threads:"
            r" (?P<task_total>[0-9]+) total,"
            r" +(?P<task_running>[0-9]+) running,"
            r" +(?P<task_sleeping>[0-9]+) sleeping,"
            r" +(?P<task_stopped>[0-9]+) stopped,"
            r" +(?P<task_zombie>[0-9]+) zombie"
        ),
        (("self.config.threads_not_tasks", "True"),),
        scope=scope,
    )

    # Cpu(s): 51.8%us, 28.7%sy,  0.5%ni, 13.9%id,  1.4%wa,  0.0%hi,  3.7%si,  0.0%st
    self._re_procps_cpu = re.compile(
        r"^%?Cpu(\(s\)|(?P<cpu_id>[0-9]+) +):"
        r" *(?P<cpu_user>[0-9.]*)[% ]us,"
        r" *(?P<cpu_system>[0-9.]+)[% ]sy,"
        r" *(?P<cpu_nice>[0-9.]+)[% ]ni,"
        r" *(?P<cpu_idle>[0-9.]+)[% ]id,"
        r" *(?P<cpu_wait>[0-9.]+)[% ]wa,"
        r" *(?P<cpu_hw_irq>[0-9.]+)[% ]hi,"
        r" *(?P<cpu_sw_irq>[0-9.]+)[% ]si,"
        r" *(?P<cpu_steal>[0-9.]+)[% ]st"
    )

    # CPU:   5% usr   3% sys   0% nic  83% idle   7% io   0% irq   0% sirq
    self._re_busybox_cpu = re.compile(
        r"^%?CPU(?P<cpu_id>[0-9]+)?:"
        r" *(?P<cpu_user>[0-9.]*)% usr"
        r" *(?P<cpu_system>[0-9.]+)% sys"
        r" *(?P<cpu_nice>[0-9.]+)% nic"
        r" *(?P<cpu_idle>[0-9.]+)% idle"
        r" *(?P<cpu_wait>[0-9.]+)% io"
        r" *(?P<cpu_hw_irq>[0-9.]+)% irq"
        r" *(?P<cpu_sw_irq>[0-9.]+)% sirq"
    )

    # Load average: 2.44 2.24 2.30 3/2473 953092
    self._re_busybox_load_average = re.compile(
        r"^%?Load average:"
        r" *(?P<load_average>[0-9.]+)"
        r" (?P<load_average_5>[0-9.]+)"
        r" (?P<load_average_15>[0-9.]+)"
        r" (?P<task_running>[0-9.]+)/(?P<task_total>[0-9]+)"
    )

    # Mem:   4046364k total,  2847408k used,  1198956k free,    37528k buffers
    self._re_procps_mem = Re_Variants(
        "re_procps_mem",
        re.compile(
            r"^Mem:"
            r" +(?P<mem_total>[0-9]+)k total,"
            r" +(?P<mem_used>[0-9]+)k used,"
            r" +(?P<mem_free>[0-9]+)k free,"
            r" +(?P<mem_buffers>[0-9]+)k buffers"
        ),
    )

    # Swap:  2047996k total,        0k used,  2047996k free,  1468792k cached
    self._re_procps_swap = Re_Variants(
        "re_procps_swap",
        re.compile(
            r"^Swap:"
            r" +(?P<swap_total>[0-9]+)k total,"
            r" +(?P<swap_used>[0-9]+)k used,"
            r" +(?P<swap_free>[0-9]+)k free,"
            r" +(?P<mem_cached>[0-9]+)k cached"
        ),
    )

    for unit in ["K", "M", "G", "T", "P", "E"]:
        # <unit>iB Mem : 15653.4 total,  6178.4 free,  7285.0 used,  2189.9 buff/cache
        self._re_procps_mem.append(
            re.compile(
                rf"^{unit}iB Mem :"
                r" +(?P<mem_total>[.0-9]+) total,"
                r" +(?P<mem_free>[.0-9]+) free,"
                r" +(?P<mem_used>[.0-9]+) used,"
                r" +(?P<mem_buffers>[.0-9]+)"
                r" buff/cache"
            ),
            (("self.mem_unit", f"{unit}iB"),),
            scope=scope,
        )

        # <unit>MiB Swap: 15792.0 total, 10146.5 free,  5645.5 used.  7242.8 avail Mem
        self._re_procps_swap.append(
            re.compile(
                rf"^{unit}iB Swap:"
                r" +(?P<swap_total>[.0-9]+) total,"
                r" +(?P<swap_free>[.0-9]+) free,"
                r" +(?P<swap_used>[.0-9]+) used\."
                r" +(?P<mem_available>[.0-9]+)"
                r" avail"
            ),
            (("self.config.mem_cached_or_available", "mem_available"),),
            scope=scope,
        )

    # 2019-01-31 06:40:41:709
    self._re_timestamp = re.compile(r"^\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d:\d\d\d$")

    self._re_process = None
    self._re_process_header = None


# ------------------------------------------------------------------------------


def _generate_re_process(
    self, columnheader_to_regex, line
) -> Tuple[str, re.Pattern, re.Pattern]:
    found = {}
    header_pattern = "^"
    process_pattern = "^"
    prespace = " *"

    line = line.rstrip("\n")

    for columnheader in re.findall(r"([^ ]+)", line):
        found[columnheader] = True
        header_pattern += prespace + columnheader.replace("+", "\\+")

        if columnheader in columnheader_to_regex:
            if columnheader_to_regex[columnheader]:
                process_pattern += prespace + columnheader_to_regex[columnheader]
                prespace = " +"
        else:
            print(
                ">INFO: header line contains unhandled columnheader"
                f" '{columnheader}'",
                file=sys.stderr,
            )
            process_pattern += prespace + r"(?:[^ ]+)"

    header_pattern += r"\s*$"
    process_pattern += r"$"

    missing = []

    for musthave in MUSTHAVE_COLUMNS[self.config.toptype]:
        if musthave not in found:
            missing.append(musthave)

    if len(missing) > 0:
        print(
            f">ERR: missing essential process column(s): {missing}\nAborting.",
            file=sys.stderr,
        )
        sys.exit(1)

    if self.config.verbosity > 2:
        print(f"header_pattern : {header_pattern}")
        print(f"process_pattern: {process_pattern}")

    return (
        header_pattern,
        re.compile(header_pattern),
        re.compile(process_pattern),
    )


# ------------------------------------------------------------------------------


def _generate_procps_re_process(self, line) -> Tuple[str, re.Pattern, re.Pattern]:
    if self.config.threads_not_tasks:
        # Command *must* be the rightmost column
        command_pattern = r"(?P<commandline>(?P<command>.+))"
    else:
        # Command shouldn't contain spaces
        command_pattern = r"(?P<commandline>(?P<command>[^ ]+).*)"

    columnheader_to_regex = {
        "%CPU": r"(?P<cpu>[\d.]+)",
        "COMMAND": command_pattern,
        "%MEM": r"(?P<mem>[\d.]+[mg]?)",
        "NI": r"(?P<nice>[\d-]+)",
        "P": r"(?P<cpu_id>\d+)",
        "PID": r"(?P<pid>[0-9]+)",
        "PR": r"(?P<priority>[\drRtT-]+)",
        "RES": r"(?P<res>[\d.]+[mg]?)",
        "S": r"(?P<state>[DINRSTtWXZ<]+)",
        "SHR": r"(?P<shr>[\d.]+[mg]?)",
        "SWAP": r"(?P<swap>[\d.]+[mg]?)",
        "TIME+": r"(?P<time>[\d:.]+)",
        "USER": r"(?P<user>[\w+-]+)",
        "VIRT": r"(?P<virt>[\d.]+[mg]?)",
    }

    return _generate_re_process(self, columnheader_to_regex, line)


# ------------------------------------------------------------------------------


def _generate_busybox_re_process(self, line) -> Tuple[str, re.Pattern, re.Pattern]:
    # PID  PPID USER     STAT   VSZ %VSZ %CPU COMMAND
    #
    # Can also have a CPU column, compile time config option
    #
    # Have I maaned about BusyBox enough yet?
    # But FFS, sometimes there's no space between the PID and PPID columns.
    # PID's min width is 5 digits, PPID's is 6.
    #
    # Some examples from running BusyBox top 1.30.1 on my laptop:
    #
    #     275     2
    #     277     2
    #   1211762     2
    #   911329     1
    #   911456911329
    #   1155184187009
    #   10164901016486

    columnheader_to_regex = {
        "CPU": r"(?P<cpu_id>[\d.]+)",
        "%CPU": r"(?P<cpu>[\d.]+)%?",
        "COMMAND": r"(?P<commandline>(?P<command>[^ ]+).*)",
        # Since topplot isn't (currently) using PPID, we can use a merged
        # PID/PPID as a single identifier. Merge can include a space. Try
        # to avoid USER starting with a digit. (Can that ever happen?)
        #
        #        (1      2     (3       (4          ))5)
        "PID": r"(?P<pid>[0-9]+( +[0-9]+(?![a-zA-Z_]))?)",
        #
        # 1 : The named capture group "pid"
        #
        # 2 : A continuous run of digits
        #
        # 3 : An unamed sub-group starting with a space followed by a
        #     continuous run of digits. Spoiler: the '?' at 5 makes this
        #     entire sub-group optional, i.e capture if present, but don't
        #     sweat if it's absent.
        #
        # 4 : A negative lookahead (non-capturing) to make 3 not match when
        #     3's run of digits is immediately followed by an alphabetic or
        #     underscore. This is in case a username starting with a digit
        #     would otherwise be captured when following a merged PID/PPID.
        #     Obvs. will still capture a username soley consiting in digits.
        #
        # Could do the following which which capture PPID when it's not
        # merged with PID, but I'm not sure that this is useful:
        #
        #   "PID": r"(?P<pid>[0-9]+( (?P<ppid>[0-9]+(?![a-zA-Z_])))?)",
        "PPID": None,
        "STAT": r"(?P<stat>[DINRSTtWXZ<]+)",
        "USER": r"(?P<user>[\w+-]+)",
        "VSZ": r"(?P<mem_abs>[\d.]+[mg]?)",
        "%VSZ": r"(?P<mem>[\d.]+)%?",
    }

    return _generate_re_process(self, columnheader_to_regex, line)


# ------------------------------------------------------------------------------



# pylint: disable=no-member

import os
import re
from typing import Callable

from .config import Config
from .dataframes import Dataframes
from .graphs import Grapher
from .parser import Parser
from .progress_window import ProgressWindow
from .toptype import TopType
from .graphmap import GraphMap, GraphMapEntry
from .utils import info

# ------------------------------------------------------------------------------


class App:
    def __init__(self, version):
        self.config = Config(self, version)

        self.setup_profiling()

        self.progress_window = (
            None
            if self.config.list_processes or not self.config.do_graph
            else ProgressWindow()
        )

    # --------------------------------------------------------------------------

    def update_status(self, msg):
        if self.config.verbosity > 0:
            print(f"status: {msg}")
        if self.progress_window is not None:
            self.progress_window.update_status(msg)

    # --------------------------------------------------------------------------

    def setup_profiling(self):
        try:
            # pylint: disable=import-outside-toplevel,unused-import
            from .profiler import Profiler
        except ImportError:
            # No profiler available, convert calls to empty stubs
            class DummyProfiler:
                def start(self, *args, **kwargs):
                    pass

                def start_new(self, *args, **kwargs):
                    pass

                def stop(self, *args, **kwargs):
                    pass

            self.profiler = DummyProfiler()
        else:
            tag1 = re.sub(  # strip everything after last '.'
                r"\.[^\.]+$", "", os.path.basename(self.config.toplog_filename)
            )

            if self.config.which_graphs is not None:
                tag1 = f"{tag1}_{self.config.which_graphs}"

            self.profiler = Profiler(
                self.config.profiling_tag is not None,
                tag1=tag1,
                tag2=self.config.profiling_tag,
            )

    # --------------------------------------------------------------------------

    def produce_graphs(
        self, dataframes: Dataframes, at_showtime: Callable, at_close: Callable
    ):
        self.update_status("producing graphs")

        cpus_graph_title = "cpu data"
        tasks_graph_title = "task data"
        poi_graph_title = f"processes of interest\n{self.config.poi_categories}"
        mem_graph_title = "mem data"

        # Awkward initialization sequence to allow graph_map to reference grapher functions
        graph_map = GraphMap()

        graphs = Grapher(
            graph_map,
            self.config,
            dataframes,
            self.config.mem_unit,
            self.config.mem_cached_or_available,
            self.progress_window,
        )

        tasks_graph = graphs.graph_procps_tasks
        if self.config.toptype == TopType.BUSYBOX:
            tasks_graph = graphs.graph_busybox_tasks

        # Note: Order here determines display order in overview figure
        graph_map[poi_graph_title] = GraphMapEntry(graphs.graph_poi, dataframes.poi_df)
        graph_map[cpus_graph_title] = GraphMapEntry(
            graphs.graph_cpus, dataframes.cpus_df
        )
        graph_map[mem_graph_title] = GraphMapEntry(graphs.graph_mem, dataframes.mem_df)
        graph_map[tasks_graph_title] = GraphMapEntry(tasks_graph, dataframes.tasks_df)

        graphs.doit(at_showtime, at_close)

    # --------------------------------------------------------------------------

    def run(self):
        self.update_status(f"parsing '{os.path.basename(self.config.toplog_filename)}'")
        self.profiler.start("parse")
        parser = Parser(self.config)
        parser.parse()

        if self.config.list_processes:
            parser.list_processes()
            self.progress_window.hide()
            self.profiler.stop()
            return

        self.update_status("munging")
        parser.munge_to_files()

        if parser.get_poi_count() < 1:
            info(
                "INFO: No processes of interest according to selection criteria:"
                f" {self.config.poi_categories}"
            )
            if self.progress_window:
                self.progress_window.hide()
            self.profiler.stop()
            return

        if self.config.do_graph:
            self.update_status("prepping dataframes")
            dataframes = parser.dataframes_from_files()
            self.profiler.start_new("graph")
            self.produce_graphs(
                dataframes, self.progress_window.hide, self.progress_window.destroy
            )

        self.profiler.stop()


# ------------------------------------------------------------------------------
# vi: sw=4:ts=4:et:tw=0



# Copyright (c) 2019-2020 Jonathan Sambrook and Codethink Ltd.
# -----------------------------------------------------------------------------
# Class to keep all instances of a plotted line having the same colour,
# linestyle, and marker across different plots.


class LineStyler:

    arbitrary_marker_offset = 0

    linestyles = ["-", ":", "-.", "--"]

    linestyle_count = len(linestyles)

    marks = [
        ".",  # point
        #    ",", # pixel
        "o",  # circle
        "v",  # triangle_down
        "^",  # triangle_up
        "<",  # triangle_left
        ">",  # triangle_right
        "1",  # tri_down
        "2",  # tri_up
        "3",  # tri_left
        "4",  # tri_right
        # "8", # octagon # Looks too much like a circle when small
        "s",  # square
        "p",  # pentagon
        "P",  # plus (filled)
        "*",  # star
        "h",  # hexagon1
        "H",  # hexagon2
        "+",  # plus
        "x",  # x
        "X",  # x (filled)
        "D",  # diamond
        "d",  # thin_diamond
        "|",  # vline
    ]

    mark_count = len(marks)

    def __init__(self):
        self.colours = {}
        self.linestyles = {}
        self.marks = {}

        self.arbitrary_marker_offset = LineStyler.arbitrary_marker_offset
        LineStyler.arbitrary_marker_offset += 5
        self.count = 0

    # -------------------------------------------------------------------------
    # Define one if undefined, then apply a line's style.

    def style(self, name, line):
        if name not in self.marks:
            self.count += 1
            self.marks[name] = LineStyler.marks[
                (self.count + self.arbitrary_marker_offset) % LineStyler.mark_count
            ]
            self.linestyles[name] = LineStyler.linestyles[
                int(self.count / LineStyler.mark_count) % LineStyler.linestyle_count
            ]
            self.colours[name] = line.get_color()

        # print(
        #    f"mark: >{self.marks[name]}<  ls: >{self.linestyles[name]}<  col:"
        #    f" >{self.colours[name]}<  name: >{name}<"
        # )

        line.set_color(self.colours[name])
        line.set_linestyle(self.linestyles[name])
        line.set_marker(self.marks[name])


# ------------------------------------------------------------------------------
# vi: sw=4:ts=4:et:tw=0



# Copyright (c) 2019-2020 Jonathan Sambrook and Codethink Ltd.
#
#    This file is part of Topplot.
#
#    Topplot is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Topplot is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Topplot.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------------

import os
import re
import sys
from threading import Timer

import matplotlib as mpl
import matplotlib.pyplot as plt


class DummyEvent:
    def __init__(self, artist, mouseevent):
        self.artist = artist
        self.mouseevent = mouseevent


# -----------------------------------------------------------------------------
# Class to abstract the management of Figs (windows)
#
# The Grapher has one or more FigManagers. It encapsulates the graphing side of topplot.
# A FigManager has one figure with one or more plots/subplots.


class FigManager:
    class PairHierarchy:
        def __init__(self):
            self._primaries = []
            self._map = {}

        def append(self, primary, secondary):
            self._primaries.append(primary)
            self._map[primary] = secondary
            self._map[secondary] = primary

        def __contains__(self, key):
            return key in self._map

        def __len__(self):
            return len(self._map)

        def __getitem__(self, index):
            if isinstance(index, int):
                len_primaries = len(self._primaries)
                if index > 2 * len_primaries:
                    raise KeyError(f"Invalid index: '{index}'")
                return (
                    self._primaries[index]
                    if index < len_primaries
                    else self._map[self._primaries[index]]
                )
            return self._map[index]

        def primary(self, key):
            if key in self._map:
                return key if self.is_primary(key) else self._map[key]
            raise KeyError(f"Not in PairHierarchy: '{key}'")

        def is_primary(self, key):
            return key in self._primaries

        def is_secondary(self, key):
            return key in self._map and not self.is_primary(key)

    def __init__(
        self,
        name,
        parent,
        title,
        window_size,
        x,
        y,
        fig=None,
        subtitle=False,
        *,
        share_legends_regex=None,
    ):
        self.name = name
        self.parent = parent

        self.ax_pairs = (
            FigManager.PairHierarchy()
        )  # Know who your buddy is, if you have one
        self.ax_name_map = {}  # Know who axes are

        self.leglines = {}  # map legend lines  to plotted lines
        self.legtexts = {}  # map legend lines  to legend texts and vice versa
        self.legmarkers = {}  # map legend lines  to legend line markers
        self.legtitles = {}  # map legend titles to plotted lines
        self.legotm = {}  # map otm legend elements to leglines from src legend
        self.legotmtitles = {}  # map otm legend title to otm leglines
        self.legotmtexts = {}  # map otm legend lines to otm legend texts and vice versa

        self.legends = {}  # map of extra legends (i.e. not ones that came with an axis)

        # Used to indicate that clicks should be replicated or shared across all
        # peering legends
        self.share_legends_regex = share_legends_regex

        self.title, _, self.subtitle = title.partition("\n")
        self.fig = plt.figure() if fig is None else fig
        self.plots = self.fig.subplots(x, y)
        self.fig.canvas.set_window_title(self.title)

        win = self.fig.canvas.manager.window
        win.geometry(window_size)  # Tkinter specific
        win.update_idletasks()
        win.after_idle(self.parent.progress_window.center, win)

        self.fig.suptitle(self.title, weight="semibold")
        if subtitle:
            # WARNING: going off mpl API here to get the figure's title's
            # position and fontsize
            x, y = self.fig._suptitle.get_position()
            fontsize = self.fig._suptitle.get_fontsize()
            self.fig.text(
                0.5, 1 - (1 - y) * 2, self.subtitle, fontsize=fontsize, ha="center"
            )

        self.fig.set_facecolor(parent.colours["fig_face"])

        #    cid = self.fig.canvas.mpl_connect('button_press_event', self.onclick)
        _ = self.fig.canvas.mpl_connect("key_press_event", self.onkey)
        _ = self.fig.canvas.mpl_connect("close_event", self.onclose)
        _ = self.fig.canvas.mpl_connect("pick_event", self.onpick)
        _ = self.fig.canvas.mpl_connect("motion_notify_event", self.onmove)

        self.help_text_box = None
        self.showing_help = 0

        self.msg_text_box = None
        self.msg_timer = None

        self.toggle_axes_legend_state = {}

        self.key_presses = {}

        self.key_presses["c"] = [
            lambda self, e: self.display_msg("No multi-core POI data available.", 1)
            if self.parent.config.cores < 2
            else self.display_msg("--dont-graph-cpu-lines on the commandline", 1)
            if not self.parent.config.plot_poi_cpu_lines
            else self.parent.graph_poi_per_cpu(use_progress_window=True),
            "Display POI data in a separate graph for each CPU core.",
        ]

        self.key_presses["C"] = [
            lambda self, e: self.parent.graph_cpu_per_cpu(use_progress_window=True)
            if self.parent.config.has_cpu_rows
            else self.display_msg("No multi-core CPU data available.", 1),
            "Display CPU data in a separate graph for each CPU core.",
        ]

        if self.parent.mplcursors_present:
            self.key_presses["e"] = [
                None,  # Documenting mplcursors default action
                "Toggle whether mplcursors is active or not.",
            ]

        self.key_presses["l"] = [
            lambda self, e: self.toggle_legends(e.inaxes),
            "Toggle legend(s). [#1]",
        ]

        self.key_presses["h"] = [
            lambda self, e: self.display_help(),
            "Cycle through the help pages.",
        ]

        self.key_presses["p"] = [
            lambda self, e: self.save(),
            "Save an image of the current window.",
        ]

        self.key_presses["P"] = [
            lambda self, e: self.parent.save_figs(),
            "Save images all the open windows.",
        ]

        self.key_presses["q"] = [
            None,  # Documenting mpl default action
            "Close the current window.",
        ]

        self.key_presses["Q"] = [
            lambda self, e: self.parent.close_figs(all_figs=True),
            "Exit topplot.",
        ]

        self.key_presses["s"] = [
            None,  # Documenting mpl default action
            "Save an image of the current window to a selectable location.",
        ]

        self.key_presses["t"] = [
            lambda self, e: self.toggle_axes(e.inaxes).fig.canvas.draw(),
            "Cycle through mem/cpu/both axis visibility for POI graph. [#1]",
        ]

        if self.parent.mplcursors_present:
            self.key_presses["v"] = [
                None,  # Documenting mplcursors default action
                "Toggle annotations' visibility. [mplcursors feature]",
            ]

        self.key_presses["0"] = [
            lambda self, e: self.parent.graph_overview(use_progress_window=True),
            "Display the overview graphs.",
        ]

        for n in range(4):

            title, _, subtitle = self.parent.ordinal_to_title(n).partition("\n")

            self.key_presses[f"{str(n+1)}"] = [
                lambda self, e: self.parent.graph_by_overview_ordinal(
                    int(e.key) - 1,
                    use_progress_window=True,
                ),
                f"Display the '{title}' graph in a separate window.",
            ]

        self.key_presses["escape"] = [
            lambda self, e: self.clear_message(),
            "Clear any displayed message.",
        ]

    # -------------------------------------------------------------------------
    # Retrieve ax's name from the map

    def ax_name(self, ax):
        return self.ax_name_map[ax] if ax in self.ax_name_map else "<unknown ax>"

    # -------------------------------------------------------------------------
    # Graphs need to be set up post-init to short circuit initialization cycles

    def setup_graphing(self, setupfn):
        if setupfn is not None:
            setupfn(self)

        plt.setp(
            [a.get_xticklabels() for a in self.fig.axes[:]],
            visible=True,
            rotation=15,
            ha="right",
        )
        self.fig.subplots_adjust(left=0.05, right=0.95, top=0.95, bottom=0.05)
        self.fig.show()

    # -------------------------------------------------------------------------
    # Given an axis and a specific legend (or set of legend lines), map the
    # legend lines to the axis' lines

    def map_legend_lines(
        self,
        ax,
        legend_title_label,
        legend=None,
        *,
        leglines=None,
        legtexts=None,
        leghandles=None,
    ):
        if legend is not None or leglines is not None:

            handles = None

            if legend is not None:
                leglines = legend.get_lines()
                if legtexts is None:
                    legtexts = legend.get_texts()
                handles = legend.legendHandles

            elif leghandles is not None:
                handles = leghandles

            titlelines = []
            titleleglines = []

            for (i, line) in enumerate(ax.get_lines()):
                if i >= len(leglines):
                    break
                legline = leglines[i]
                legtext = legtexts[i] if legtexts is not None else None

                if legline in self.leglines:
                    print(f"CLASH: legline {legline} already present")

                self.leglines[legline] = line
                legline.set_picker(5)

                if legtext is not None:
                    self.leglines[legtext] = line
                    legtext.set_picker(10)
                    self.legtexts[legline] = legtext
                    self.legtexts[legtext] = legline

                if handles is not None:
                    # pylint: disable=protected-access
                    self.legmarkers[legline] = handles[i]

                titleleglines.append(legline)
                titlelines.append(line)

            if legend_title_label in self.legtitles:
                self.legtitles[legend_title_label]["lines"] += titlelines
                self.legtitles[legend_title_label]["leglines"] += titleleglines
            else:
                self.legtitles[legend_title_label] = {
                    "leglines": titleleglines,
                    "lines": titlelines,
                }
                legend_title_label.set_picker(10)

        else:
            # raise Exception("Must have one of legend or leglines")
            print("Must have one of legend or leglines")

    # -------------------------------------------------------------------------
    # Display a message box. Optionally set a minimum (!) display period.

    def display_msg(self, txt, timeout=None):
        if self.msg_timer is not None:
            self.msg_timer.cancel()
            self.msg_timer = None

        txt += "\n\n[press Esc to clear this message]"

        if self.msg_text_box is None:
            bbox_props = dict(boxstyle="round", pad=2, facecolor="wheat", alpha=0.95)

            props = dict(fontfamily="monospace", fontsize=14, ha="center")

            self.msg_text_box = self.fig.text(0.5, 0.5, txt, bbox=bbox_props, **props)

        else:
            self.msg_text_box._text = txt
            self.msg_text_box.set_visible(True)

        self.fig.canvas.draw()

        if timeout is not None:

            def clearup():
                self.msg_text_box.set_visible(False)

            self.msg_timer = Timer(timeout, clearup)
            self.msg_timer.start()

    # -------------------------------------------------------------------------
    # Tidy away any previously displayed message

    def clear_message(self):
        if self.msg_text_box is not None:
            self.msg_text_box.set_visible(False)

        self.showing_help = 0
        self.fig.canvas.draw()

    # -------------------------------------------------------------------------
    # Display info about using topplot. Mainly runtime keyboard options.

    def display_help(self):
        def get_text(page):
            txt = ""
            if page == 1:
                txt = "Key press menu:\n\n"
                for (key, array) in self.key_presses.items():
                    item_text = array[1]
                    if item_text is not None:
                        txt += f"    '{key}' : {item_text}\n\n"

                txt += "#1: If the mouse pointer is over a graph, this applies to just that graph;"
                txt += (
                    "    otherwise it applies to all the graphs in the active window."
                )

            elif page == 2:
                txt = """Graph legends:

    Click on a legend line to hide it on the graph.
    Click on it again to display the line on the graph once more.

    Click on a legend's title to toggle the display status of each of the lines.
    I.e. if they're hidden, display them, if they're displayed, hide them.

    Right click on a legend's title to reset all the lines to being displayed.

    Legends are draggable, but do not leave them entirely over a different graph.
    If this happens they lose all interactivity, including draggability!
"""

                if self.parent.mplcursors_present:
                    txt += """
Annotations:

    Clicking on plotted lines will display information at that position. Right
    click on the info to dispel it."""
            else:
                return None

            txt += f"""

Page {page}   Version: {self.parent.config.version}    """
            txt += "Copyright (c) 2019-2021 Jonathan Sambrook and Codethink Ltd."
            return txt

        # ----------------------------------------------------------------------

        if self.help_text_box is None:
            self.showing_help = 1
            bbox_props = dict(  # boxstyle='circle',
                pad=15, facecolor="white", alpha=0.975
            )

            props = dict(fontfamily="monospace", fontsize=10, ha="left")

            self.help_text_box = self.fig.text(
                0.015, 0.03, get_text(1), bbox=bbox_props, **props
            )
        else:
            next_page = self.showing_help + 1
            txt = get_text(next_page)
            if txt:
                self.showing_help = next_page
                if self.help_text_box is not None:
                    self.help_text_box.set_text(txt)
                    self.help_text_box.set_visible(True)

            else:
                self.showing_help = 0
                if self.help_text_box is not None:
                    self.help_text_box.set_visible(False)

        self.fig.canvas.draw()

    # -------------------------------------------------------------------------
    # Hook for keypresses

    def onkey(self, e):
        # print(f"you pressed: '{e.key}'   x:'{e.xdata}'   y:'{e.ydata}'     e: {e}")

        if self.showing_help and e.key != "q":
            self.help_text_box.set_visible(False)
            self.fig.canvas.draw()

        if e.key in self.key_presses and self.key_presses[e.key][0] is not None:
            self.key_presses[e.key][0](self, e)

    # -------------------------------------------------------------------------
    # Hook for mouse clicks

    def onclick(self, e):  # pylint: disable=no-self-use
        # print(
        #     "%s click: button=%d, x=%d, y=%d, xdata=%f, ydata=%f"
        #     % (
        #         "double" if e.dblclick else "single",
        #         e.button,
        #         e.x,
        #         e.y,
        #         e.xdata,
        #         e.ydata,
        #     )
        # )
        if e.inaxes is not None:
            print(f"Click in axis: '{e.inaxes.get_title()}'")

    # -------------------------------------------------------------------------
    # Hook for window closing

    def onclose(self, _):
        if self.name in self.parent.fig_manager:
            del self.parent.fig_manager[self.name]
        self.parent.close_check()

    # -------------------------------------------------------------------------
    # Hook for motion
    # This is intended to resolve occasional flickering issues without too much
    # overhead. Flicker observed on Linux. Seems likely TkInter related.

    onmove_counter = 0

    def onmove(self, _):
        self.onmove_counter += 1
        if self.onmove_counter % 100 == 0:
            self.fig.canvas.draw_idle()

    # -------------------------------------------------------------------------
    # Hook for artist picking

    def onpick(self, e):
        if isinstance(e.artist, mpl.legend.Legend):
            # Not sure how/why these occur, but we're not interested
            return

        if self.share_legends_regex:
            if e.artist in self.legtitles:
                self.shared_legend_title_pick(e)
            else:
                self.shared_legend_line_pick(e)
        else:
            if e.artist in self.legtitles:
                self.on_legend_title_pick(e)
            elif e.artist in self.legotm:
                self.on_otm_legend_line_pick(e)
            elif e.artist in self.legotmtitles:
                self.on_otm_legend_title_pick(e)
            else:
                self.on_legend_line_pick(e)

    # -------------------------------------------------------------------------
    # "Print" this fig to a file

    def save(self):
        directory = self.parent.config.output_dir
        directory = (
            ""
            if directory is None
            else f"{directory}{os.sep}"
            if directory[-1] != os.sep
            else directory
        )

        sep = "\\\\" if os.sep == "\\" else os.sep
        pattern = f"[^0-9a-zA-Z{'' if directory else sep}]+"

        title = self.title.partition("\n")[0]
        title = re.sub(pattern, "_", title)
        title = title.partition("topplot_")[2]

        filename = f"{directory}{title}.png"
        self.fig.savefig(filename)

    # -------------------------------------------------------------------------
    # Set alpha for line, marker, and text label
    # Handles either the line being clicked or the text label since self.legtexts
    # map is bi-directional

    def set_legline_alpha(self, legline, alpha):
        legline.set_alpha(alpha)

        if legline in self.legtexts:
            self.legtexts[legline].set_alpha(alpha)

        if legline in self.legmarkers:
            self.legmarkers[legline].set_alpha(alpha)

        elif self.legtexts[legline] in self.legmarkers:
            self.legmarkers[self.legtexts[legline]].set_alpha(alpha)

    # -------------------------------------------------------------------------
    # Hook for legend title clicks

    def on_legend_title_pick(self, e1, *, redraw=True):
        # on the pick event, find the orig line corresponding to the
        # legend proxy line, and toggle the visibility
        pickee = e1.artist

        rmb = e1.mouseevent.button == mpl.backend_bases.MouseButton.RIGHT
        shift = e1.mouseevent.key == "shift"

        for (i, line) in enumerate(self.legtitles[pickee]["lines"]):
            if rmb or shift:
                legvis = True
            else:
                legvis = not line.get_visible()

            line.set_visible(legvis)

            # Change the alpha on the line in the legend so we can see which lines
            # have been toggled
            legline = self.legtitles[pickee]["leglines"][i]
            self.set_legline_alpha(legline, 1.0 if legvis else 0.2)

        if redraw:
            self.fig.canvas.draw_idle()

    # -------------------------------------------------------------------------
    # Hook for legend title clicks with shared legends
    # Note that this doesn't distinguish between legends. Any line in any legend
    # may match.

    def shared_legend_title_pick(self, e1):
        text = e1.artist._text
        for legend_title in self.legtitles:
            if legend_title._text == text:
                e2 = DummyEvent(legend_title, e1.mouseevent)
                self.on_legend_title_pick(e2, redraw=False)

        self.fig.canvas.draw_idle()

    # -------------------------------------------------------------------------
    # Hook for legend line clicks

    def on_legend_line_pick(self, e1, *, redraw=True):
        # on the pick event, find the orig line corresponding to the
        # legend proxy line, and toggle the visibility
        legline = e1.artist

        rmb = e1.mouseevent.button == mpl.backend_bases.MouseButton.RIGHT
        shift = e1.mouseevent.key == "shift"
        force_visible = rmb or shift

        if legline in self.leglines:
            origline = self.leglines[legline]
            vis = True if force_visible else not origline.get_visible()
            origline.set_visible(vis)

            # Change the alpha on the line in the legend so we can see what lines
            # have been toggled
            self.set_legline_alpha(legline, 1.0 if vis else 0.2)

            if redraw:
                self.fig.canvas.draw_idle()

    # -------------------------------------------------------------------------
    # Hook for legend line clicks with shared legends

    def shared_legend_line_pick(self, e1, *, redraw=True):
        # Sort out legline and label : either could've been clicked to get here
        # and we need the label to "click" all the peers in the shared legends.
        if isinstance(e1.artist, mpl.lines.Line2D):
            legline = e1.artist
            label = self.legtexts[legline]
        else:
            label = e1.artist
            legline = self.legtexts[label]

        label_text = label._text

        # Use regex because the label might need picking apart to extract
        # partial content for comparison.
        re_category = re.compile(self.share_legends_regex)

        m = re_category.match(label_text)
        category = m.group(1)

        for line, text in self.legtexts.items():
            m = re_category.match(text._label)
            if m and m.group(1) == category:
                e2 = DummyEvent(line, e1.mouseevent)
                self.on_legend_line_pick(e2, redraw=False)

        if redraw:
            self.fig.canvas.draw_idle()

    # -------------------------------------------------------------------------
    # Hook for otm legend line clicks
    # Work actually done by the "real" legend's members.

    def on_otm_legend_line_pick(self, e1, *, redraw=True):
        if isinstance(e1.artist, mpl.lines.Line2D):
            legline = e1.artist
            label = self.legotmtexts[legline]
        else:
            label = e1.artist
            legline = self.legotmtexts[label]

        rmb = e1.mouseevent.button == mpl.backend_bases.MouseButton.RIGHT
        shift = e1.mouseevent.key == "shift"
        force_visible = rmb or shift

        # De-reference the legend's element to the "real" legend's lines
        for otmlegline in self.legotm[legline]:
            e2 = DummyEvent(otmlegline, e1.mouseevent)
            self.on_legend_line_pick(e2, redraw=False)

        # Toggle the otm legend line and label
        if force_visible:
            alpha = 999
        else:
            alpha = legline.get_alpha()

        # alpha is None (implying 1.0) if it's not already been explicitly set
        alpha = 0.2 if (alpha is None or alpha == 1.0) else 1.0

        legline.set_alpha(alpha)
        label.set_alpha(alpha)

        if redraw:
            self.fig.canvas.draw_idle()

    # -------------------------------------------------------------------------
    # Hook for otm legend title clicks

    def on_otm_legend_title_pick(self, e1):
        otm_title = e1.artist

        for otm_legline in self.legotmtitles[otm_title]:
            e2 = DummyEvent(otm_legline, e1.mouseevent)
            self.on_otm_legend_line_pick(e2, redraw=False)

        self.fig.canvas.draw_idle()

    # -------------------------------------------------------------------------
    # Get the window to the top of the z-order

    def show(self):
        manager = self.fig.canvas.manager

        # Tkinter specific code follows. If you change GUI toolkit (e.g. to
        # QT), this will break.
        window = manager.window

        # .lift() and .tkraise() should but don't work. The following is ugly but does.
        window.attributes("-topmost", 1)

        if not sys.platform.startswith("win"):
            # Necessary on my Linux WM, but not on Windows
            window.withdraw()
            window.deiconify()

        window.update_idletasks()

        window.attributes("-topmost", 0)

    # -------------------------------------------------------------------------
    # Close the window

    def close(self):
        plt.close(self.fig)

        # Tidy up
        self.onclose(None)

    # -------------------------------------------------------------------------
    # Short cut to return the wrapped Fig's axes

    def get_axes(self):
        return self.fig.get_axes()

    # -------------------------------------------------------------------------
    # Track legend by axis and name

    def register_legend(self, ax, name, legend):
        if ax not in self.legends:
            self.legends[ax] = {}
        if name not in self.legends[ax]:
            self.legends[ax][name] = {}
        self.legends[ax][name]["legend"] = legend

    # -------------------------------------------------------------------------
    # Retrieve the legend of the given name if it has been registered with the given axis.

    def named_legend(self, ax, name):
        if ax in self.legends and name in self.legends[ax]:
            return self.legends[ax][name]["legend"]
        return None

    # -------------------------------------------------------------------------
    # Retrieve the Legends of the given axis, and its pair, if one's been
    # given; otherwise return Legends
    # from all of the Fig's axes

    def get_legends(self, ax=None):
        if ax is None:
            axes = self.get_axes()
        else:
            axes = [ax]
            if ax in self.ax_pairs:
                axes.append(self.ax_pairs[ax])

        legends = []
        for ax in axes:
            for c in ax.get_children():
                if isinstance(c, mpl.legend.Legend):
                    legends.append(c)

        return legends

    # -------------------------------------------------------------------------
    # Toggle the visibility of the Legends associated with an axis and any pair it has

    def toggle_legends(self, ax=None):
        if ax is None:
            legends = self.get_legends()
        else:
            legends = self.get_legends(ax)

        visibility = False

        for legend in legends:
            visibility = legend.get_visible()
            if visibility:
                break

        for legend in legends:
            legend.set_visible(not visibility)

        self.fig.canvas.draw()

    # -------------------------------------------------------------------------
    # Cycle through one, then t'other, then back to both of the axes being visible.

    def toggle_axes(self, ax):
        if ax is None:
            axes = []
            for ax1 in self.ax_pairs:
                if self.ax_pairs[ax1] not in axes:
                    axes.append(ax1)

        else:
            axes = [self.ax_pairs.primary(ax) if ax in self.ax_pairs else ax]

        for ax1 in axes:
            if ax1 in self.ax_pairs:
                ax2 = self.ax_pairs[ax1]

                # print(f"ax1: {hex(id(ax1))}  ax2: {hex(id(ax2))}")

                # If an axis has the other axis' legend attached to it for
                # z-order reasons, this is the way to get hold of it.
                cuckoo = self.named_legend(ax1, "cuckoo")
                if cuckoo is None:
                    cuckoo = self.named_legend(ax2, "cuckoo")

                v1 = ax1.get_visible()
                v2 = ax2.get_visible()
                # print(f"v1: {v1}  v2: {v2}")
                if v1 and v2:
                    # print(f"ax2/{self.ax_name(ax2)} only")
                    ax1.set_visible(False)

                    if cuckoo is not None:
                        # Transfer cuckoo between axes
                        cuckoo.remove()
                        ax2.add_artist(cuckoo)

                    legend2 = ax2.get_legend()
                    if legend2 is not None and not legend2.get_visible():
                        legend2.set_visible(True)
                        self.toggle_axes_legend_state[ax2] = legend2

                elif (not v1) and v2:
                    # print(f"ax1/{self.ax_name(ax1)} only")
                    ax1.set_visible(True)
                    ax2.set_visible(False)

                    # twinx sets the secondary axis's labels, ticks, grid and
                    # background patch to invisible
                    ax1.patch.set_visible(True)
                    ax1.get_xaxis().set_visible(True)  # Covers labels, ticks and grid

                    if cuckoo is not None and cuckoo.get_visible():
                        cuckoo.remove()
                        ax1.add_artist(cuckoo)
                        cuckoo.set_visible(False)
                        self.toggle_axes_legend_state[ax1] = cuckoo

                    if ax2 in self.toggle_axes_legend_state:
                        self.toggle_axes_legend_state[ax2].set_visible(False)
                        del self.toggle_axes_legend_state[ax2]

                else:
                    # print("both axen")
                    ax2.set_visible(True)

                    # twinx sets the secondary axis's labels, ticks, grid and
                    # background patch to invisible
                    ax1.patch.set_visible(False)
                    ax1.get_xaxis().set_visible(False)  # Covers labels, ticks and grid

                    if cuckoo is not None:
                        cuckoo.set_visible(True)

                    if ax1 in self.toggle_axes_legend_state:
                        self.toggle_axes_legend_state[ax1].set_visible(True)
                        del self.toggle_axes_legend_state[ax1]

        return self


# -----------------------------------------------------------------------------
# vi: sw=4:ts=4:et



from typing import Callable, NamedTuple

from matplotlib.axis import Axis
from pandas import DataFrame

from .figman import FigManager

# ------------------------------------------------------------------------------


class GraphMapEntry(NamedTuple):
    fn: Callable[[FigManager, DataFrame, Axis, str, bool], None]
    data: DataFrame


GraphMap = dict[str, GraphMapEntry]


# ------------------------------------------------------------------------------



import pkg_resources

from .app import App


def main():
    try:
        version = pkg_resources.require("topplot")[0].version
    except pkg_resources.DistributionNotFound:
        version = "unknown (dev?)"

    topplot = App(version=version)
    topplot.run()


if __name__ == "__main__":
    main()



from pandas import DataFrame


class Dataframes:
    def __init__(self):
        self.cpus_df: DataFrame = DataFrame()
        self.tasks_df: DataFrame = DataFrame()
        self.poi_df: DataFrame = DataFrame()
        self.mem_df: DataFrame = DataFrame()
        self.core_dfs: DataFrame = DataFrame()



import matplotlib as mpl

# -----------------------------------------------------------------------------


def override_rcParams():
    """Customize MatPlotLib rcParams for topplot."""
    mpl.rcParams["legend.facecolor"] = "white"  # Doesn't work
    mpl.rcParams["legend.fancybox"] = True
    mpl.rcParams["legend.shadow"] = True
    # Disable keystrokes for toggling log scales
    mpl.rcParams["keymap.xscale"] = []
    mpl.rcParams["keymap.yscale"] = []
    # Remove clashes with topplot's keymap
    mpl.rcParams["keymap.back"] = filter(
        lambda x: x != "c", mpl.rcParams["keymap.back"]
    )


# -----------------------------------------------------------------------------


def rcParams():
    """Dump MatPlotLib rcParams."""
    # Override defaults
    override_rcParams()

    # Dump state
    for k, v in mpl.rcParams.items():
        print(f"mpl.rcParams['{k}'] = {v}")


# -----------------------------------------------------------------------------



from enum import Enum


class TopNDict2:
    class AppendType(Enum):
        APPEND_LAZY = 1
        APPEND_SORT_AND_SETTLE = 2
        APPEND_BISECT_INSERT = 3

    def __init__(
        self,
        n: int,
        name: str,
        dictionary: dict = None,
        append_type: AppendType = AppendType.APPEND_BISECT_INSERT,
        verbosity: int = 0,
    ):
        """Generate a two level dictionary of the top N appended items.

        Can take an active or lazy approach to appending items.

        Active means sorting and trimming the list on every append. This keeps
        the size down, at the expense of more sorts being performed, albeit on a
        list of fixed maximum size.

        Lazy adds items to the list, deferring sorting and trimming until
        completion is triggered. Less upfront work, more space may be used, final
        sort will be on all added items.

        Args:

            n (int): Maximum number of entries in the dictionary.

            name (str): Name of this instance.

            dictionary (dict, optional): Where the entries end up. If none is
                specified a new one is created to be returned at completion.

            lazy (bool, optional): Use the lazy append or not. Defaults to True.

            verbosity (int, optional): Display info on stdout. Defaults to 0.
        """
        self._n = n
        self.name = name
        self._dict = {} if dictionary is None else dictionary
        self._append_type = append_type
        self._verbosity = verbosity

        self._needs_settling = False
        self._list = [("", "", 0)]

    # --------------------------------------------------------------------------

    def settle(self):
        """Sort and trim the internal list as necessary"""

        if self._needs_settling:
            self._needs_settling = False
            self._list.sort(key=take_third, reverse=True)

            if len(self._list) > self._n:
                self._list = self._list[0 : self._n]

    # --------------------------------------------------------------------------

    def append_lazily(self, key1, key2, value):
        """Append the specified item, deferring sorting and trimming.

        Args:
            key1 : Key to first level of dict stack
            key2 : Key to second level of dict stack
            value : Value of item
        """

        self._needs_settling = True
        self._list.append((key1, key2, value))

    # --------------------------------------------------------------------------

    def append_and_settle(self, key1, key2, value):
        """Append the specified item if it meets the current criteria.

        Append if either the list isn't yet full, or if the new value
        supercedes that of (at least) the current tail.

        Args:
            key1 : Key to first level of dict stack
            key2 : Key to second level of dict stack
            value : Value of item
        """

        if self._needs_settling:
            self.settle()

        if (len(self._list) < self._n) or (value > self._list[-1][2]):
            self._list.append((key1, key2, value))
            self.settle()

    # --------------------------------------------------------------------------

    def bisect_insert(self, key1, key2, value):
        """Insert the specified item if it meets the current criteria.

        Insert if either the list isn't yet full, or if the new value
        supercedes that of (at least) the current tail.

        Args:
            key1 : Key to first level of dict stack
            key2 : Key to second level of dict stack
            value : Value of item
        """

        if self._needs_settling:
            self.settle()

        if (len(self._list) < self._n) or (value > self._list[-1][2]):

            # The list is known to be sorted, so bisect insertion should
            # be cheaper than append-then-sort.

            lo = 0
            hi = len(self._list)

            while lo < hi:
                mid = (lo + hi) // 2
                if value > self._list[mid][2]:
                    hi = mid
                else:
                    lo = mid + 1

            self._list.insert(lo, (key1, key2, value))

            if len(self._list) > self._n:
                self._list = self._list[0 : self._n]

    # --------------------------------------------------------------------------

    def append(self, key1, key2, value, append_type: AppendType = None):
        """Handle appending the specified item to the list.

        If 'lazy', either according to the instance's setting or, if set, this
        method's parameter, append the specified item, deferring sorting and
        trimming.

        Otherwise append the specified item if it meets the current criteria.
        I.e. append if either the list isn't yet full, or if the new value
        supercedes that of (at least) the current tail.

        Args:
            key1 : Key to first level of dict stack
            key2 : Key to second level of dict stack
            value : Value of item
            lazy (bool, optional) : Override this instance's lazyness
        """

        if append_type == self.AppendType.APPEND_LAZY:
            self.append_lazily(key1, key2, value)
        elif append_type == self.AppendType.APPEND_SORT_AND_SETTLE:
            self.append_and_settle(key1, key2, value)
        else:
            self.bisect_insert(key1, key2, value)

    # --------------------------------------------------------------------------

    def complete(self) -> dict:
        """Wrap things up and update dictionary

        Returns:
            dict: The values stored in a dictionary
        """
        if self._needs_settling:
            self.settle()

        if self._verbosity > 0:
            print(f"Top {self._n} by {self.name}:")

        for key1, key2, value in self._list:
            if key1 == "":
                print(f"WARN: Missing key1 in TopNDict2 '{self.name}'(?)")
            else:
                if key1 not in self._dict:
                    self._dict[key1] = {}

                self._dict[key1][key2] = True

                if self._verbosity > 0:
                    print(f"  {key1} [{key2}] {round(value,2)}")

        return self._dict


# ------------------------------------------------------------------------------
# Used to select third element of a tuple
def take_third(item):
    return item[2]



from ..dataframes import Dataframes
from ..tempfile import (
    CpuFiles,
    dataframe_from_process_graph_data,
    from_csv,
    write_datafile as write_temp_datafile,
)
from .topndict import TopNDict2
from ..utils import warn

# ------------------------------------------------------------------------------

CPU_COLUMNS = ["user", "system", "nice", "idle", "wait", "hw_irq", "sw_irq"]

# ------------------------------------------------------------------------------


def _write_datafile(self, filename, source, keys):
    """Write data if allowed by config."""
    if not self.config.allow_write:
        return
    write_temp_datafile(filename, source, keys)


# ------------------------------------------------------------------------------


def munge_to_files(self):
    """Convert internal data structures to temp files according to config."""
    _munge_header_info(self)
    _munge_poi_info(self)
    _munge_poi_data_to_files(self)


# ------------------------------------------------------------------------------


def _munge_header_info(self):
    """Munge top's header info to temp files."""
    if self._highest_cpuid:
        self.config.cores = int(self._highest_cpuid) + 1

    if self.config.has_cpu_rows:
        cpu_keys = ["load_average"]
        column_index = 3
        columns = CPU_COLUMNS
        if self.config.with_cpu_steal:
            columns.append("steal")
        for i in range(0, self.config.cores):
            for column in columns:
                cpu_keys.append(f"cpu{i}_{column}")
                column_index += 1
    else:
        cpu_keys = [
            "load_average",
            "cpu_user",
            "cpu_system",
            "cpu_nice",
            "cpu_idle",
            "cpu_wait",
            "cpu_hw_irq",
            "cpu_sw_irq",
        ]
        if self.config.with_cpu_steal:
            cpu_keys.append("steal")

    _write_datafile(self, self._cpu_data_filename, self._top_entries, cpu_keys)

    mem_keys = [
        "mem_used",
        "mem_free",
        "mem_buffers",
        self.config.mem_cached_or_available,
    ]

    if self.config.has_swap:
        mem_keys.append("swap_free")

    _write_datafile(self, self._mem_data_filename, self._top_entries, mem_keys)

    task_keys = ["task_running"]
    if self.config.has_full_task_info:
        task_keys += ["task_sleeping", "task_stopped", "task_zombie"]

    _write_datafile(self, self._task_data_filename, self._top_entries, task_keys)


# ------------------------------------------------------------------------------


def _munge_poi_info(self):
    """Munge top's poi info to temp data structures."""
    acc_cpu = None
    acc_mem = None
    peak_cpu = None
    peak_mem = None

    # Keep track of top N lists on request
    if self.config.poi_acc_cpu:
        acc_cpu = TopNDict2(
            self.config.poi_acc_cpu,
            "accumlated cpu",
            self._processes_of_interest,
            verbosity=self.config.verbosity,
        )

    if self.config.poi_acc_mem:
        acc_mem = TopNDict2(
            self.config.poi_acc_mem,
            "accumlated mem",
            self._processes_of_interest,
            verbosity=self.config.verbosity,
        )

    if self.config.poi_peak_cpu:
        peak_cpu = TopNDict2(
            self.config.poi_peak_cpu,
            "peak cpu",
            self._processes_of_interest,
            verbosity=self.config.verbosity,
        )

    if self.config.poi_peak_mem:
        peak_mem = TopNDict2(
            self.config.poi_peak_mem,
            "peak mem",
            self._processes_of_interest,
            verbosity=self.config.verbosity,
        )

    # Loop over all processes, keeping tabs on top N lists and filtering for POI
    for command in self._processes:  # pylint: disable=consider-using-dict-items
        for pid in self._processes[command]:
            # Update top-N lists if required
            if acc_cpu:
                acc_cpu.append(command, pid, self._processes[command][pid]["acc_cpu"])

            if acc_mem:
                acc_mem.append(command, pid, self._processes[command][pid]["acc_mem"])

            if peak_cpu:
                peak_cpu.append(command, pid, self._processes[command][pid]["max_cpu"])

            if peak_mem:
                peak_mem.append(command, pid, self._processes[command][pid]["max_mem"])

            # Run main filters
            for timestamp in self._processes[command][pid]["timestamps"].keys():
                if self.config.filterfoo(command, pid, timestamp):
                    if command not in self._processes_of_interest:
                        if command == "":
                            warn(
                                f"Adding empty command (?) for pid {pid} at"
                                f" {timestamp}"
                            )
                        self._processes_of_interest[command] = {}

                    self._processes_of_interest[command][pid] = True
                    continue

    # Extract POI from top-N lists if required. Also dump info to stdout on request
    if self.config.poi_acc_cpu:
        acc_cpu.complete()

    if self.config.poi_acc_mem:
        acc_mem.complete()

    if self.config.poi_peak_cpu:
        peak_cpu.complete()

    if self.config.poi_peak_mem:
        peak_mem.complete()


# ------------------------------------------------------------------------------


def _munge_poi_data_to_files(self):
    """Munge data for processes of interest to temp file(s)"""

    data_by_core = CpuFiles(self.config, self.tmpdir, self._poi_combined_data_filename)

    with open(self._poi_data_filename, "w", encoding="ISO-8859-1") as poi_data:
        keys = ["cpu", "mem"]

        # Data
        block_index = -1

        for (
            command
        ) in self._processes_of_interest:  # pylint: disable=consider-using-dict-items
            if command == "":
                warn("Skipping empty command (?)")
                continue

            pids = self._processes_of_interest[command]

            for pid in pids:
                if not self._processes[command][pid]["timestamps"]:
                    warn(f"Unexpected lack of timestamps for {command}:{pid}")
                    continue

                block_index += 1

                if len(pids) > 1:
                    qualified_command = f"{command}:{pid}"
                else:
                    qualified_command = f"{command}"

                # Header
                line = "timestamp"
                for key in keys:
                    key = key.replace("_", " ")
                    line = f'{line} "{qualified_command} - {key}"'

                header = f"\n{line}\n"
                poi_data.write(header)

                # Data
                for timestamp in self._processes[command][pid]["timestamps"].keys():
                    line = f"{timestamp}"
                    for key in keys:
                        line = (
                            f"{line}"
                            f" {self._processes[command][pid]['timestamps'][timestamp][key]}"
                        )
                    txt = line + "\n"
                    poi_data.write(txt)
                    if self.config.has_cpu_column:
                        core = int(
                            self._processes[command][pid]["timestamps"][timestamp][
                                "cpu_id"
                            ]
                        )

                        # Write data to split out file
                        data_by_core.write(core, header, txt)

                        # Update per core aggregate pile(s) as appropriate
                        if self.config.plot_poi_cpu_sum:
                            data_by_core.add_poi_cpu(
                                core,
                                timestamp,
                                float(
                                    self._processes[command][pid]["timestamps"][
                                        timestamp
                                    ]["cpu"]
                                ),
                            )

                        if self.config.plot_poi_mem_sum:
                            data_by_core.add_poi_mem(
                                core,
                                timestamp,
                                self._processes[command][pid]["timestamps"][timestamp][
                                    "mem"
                                ],
                            )

                poi_data.write("\n")

                if self.config.has_cpu_column:
                    data_by_core.seal_register()

        data_by_core.close()


# ------------------------------------------------------------------------------


def dataframes_from_files(self) -> Dataframes:
    """Prepare dataframes from previously munged temp files."""
    dataframes = Dataframes()

    dataframes.cpus_df = from_csv(self._cpu_data_filename)
    dataframes.tasks_df = from_csv(self._task_data_filename)
    dataframes.poi_df = dataframe_from_process_graph_data(self._poi_data_filename)
    dataframes.mem_df = from_csv(self._mem_data_filename)

    # Add extra column(s) for summary CPU "exec" info
    ns = range(self.config.cores) if self.config.has_cpu_rows else [""]
    for n in ns:
        dataframes.cpus_df[f"cpu{n} exec"] = (
            dataframes.cpus_df[f"cpu{n} system"]
            + dataframes.cpus_df[f"cpu{n} user"]
            + dataframes.cpus_df[f"cpu{n} nice"]
        )

    # Prep the poi-by-core data
    dataframes.core_dfs = []
    if self.config.has_cpu_column and self.config.cores > 1:
        for core in range(self.config.cores):
            dataframes.core_dfs.append(
                dataframe_from_process_graph_data(
                    f"{self.tmpdir}/cpu{core}_process.data"
                )
            )

    return dataframes


# ------------------------------------------------------------------------------



from io import IOBase
import re
import os
import pathlib
import tempfile
from typing import Callable
import sys

from ..config import Config
from ..toptype import TopType
from ..utils import dhms_to_sec

# ------------------------------------------------------------------------------


class Parser:
    """Parses top logs, munges the data according to config, emits dataframes."""

    # pylint: disable=import-outside-toplevel
    from ._list_processes import list_processes
    from ._munge import (
        munge_to_files,
        dataframes_from_files,
    )
    from ._regex import (
        _setup_regex,
        _generate_procps_re_process,
        _generate_busybox_re_process,
    )

    # --------------------------------------------------------------------------

    class ParseContext:
        """Intermediary data shuttle."""

        def __init__(self):
            self.cpu_id = None
            self.current_entry = {}
            self.line_count = 0
            self.batch_count = 0
            self.first_secs = None
            self.prev_secs = None
            self.re_process = None

    # --------------------------------------------------------------------------

    def __init__(self, config: Config):
        self.config = config

        self._highest_cpuid = None
        self._processes = {}
        self._processes_of_interest = {}
        self._top_entries = []

        self._re_busybox_cpu = None
        self._re_busybox_load_average = None

        self._re_procps_cpu = None
        self._re_procps_mem = None
        self._re_procps_swap = None
        self._re_procps_tasks = None

        self._re_ignore_processes_regex = None
        self._re_process_header = None
        self._re_top_busybox = None
        self._re_top_procps = None

        self._setup_regex()
        self._setup_tmpdir_and_files()

    # --------------------------------------------------------------------------

    def get_poi_count(self) -> int:
        """Return count of processes of interest."""
        return len(self._processes_of_interest)

    # --------------------------------------------------------------------------

    def _setup_tmpdir_and_files(self) -> None:
        """Prepare for the temporary data files."""
        if self.config.tmpdir is None:
            self.tmpdir_context_manager = tempfile.TemporaryDirectory()
            self.tmpdir = self.tmpdir_context_manager.name
        else:
            self.tmpdir = self.config.tmpdir
            try:
                pathlib.Path(self.tmpdir).mkdir(mode=0o700, parents=True, exist_ok=True)
            except PermissionError as e:
                print(f"Can't create self.tmpdir: {e}")
                sys.exit(1)

        if self.config.verbosity > 0:
            print(f"tmpdir: {self.tmpdir}")

        self._cpu_data_filename = os.path.join(self.tmpdir, "cpu.data")
        self._mem_data_filename = os.path.join(self.tmpdir, "mem.data")
        self._task_data_filename = os.path.join(self.tmpdir, "task.data")
        self._poi_combined_data_filename = os.path.join(
            self.tmpdir, "poi combined.data"
        )
        self._poi_data_filename = os.path.join(self.tmpdir, "poi.data")

    # --------------------------------------------------------------------------

    def parse(self) -> None:
        """Extract info from the top log file.

        Note: This only extracts the data in to internal data structures.
              Further munging in to temporary files, and then to dataframes
              is needed for graphing."""

        context = Parser.ParseContext()

        re_first_line = None
        parse_header: Callable[[IOBase, Parser.ParseContext, re.Match], None] = None

        # Encoding set to ISO-8859-1 in response to 0xFF char used in a threadname (!)
        # causing topplot to die horribly. Will this have knock on effects in non-ASCII
        # environments?

        with open(self.config.toplog_filename, encoding="ISO-8859-1") as top_log:

            for (
                line
            ) in top_log:  # Swapping over to .readlines(10000): leads to Bad Things
                context.line_count += 1
                line = line.rstrip()
                if not line:
                    continue

                # print(f"{context.line_count}: len: {len(line)}", file=sys.stderr)

                line_handled = False

                if not re_first_line:
                    top_procps_line_match = self._re_top_procps.match(line)
                    if top_procps_line_match:
                        re_first_line = self._re_top_procps
                        parse_header = self._parse_procps_header
                        self.config.toptype = TopType.PROCPS
                        self.config.has_swap = True
                        self.config.has_full_task_info = True
                    else:
                        top_busybox_line_match = self._re_top_busybox.match(line)
                        if top_busybox_line_match:
                            re_first_line = self._re_top_busybox
                            parse_header = self._parse_busybox_header
                            self.config.toptype = TopType.BUSYBOX
                            self._highest_cpuid = 0

                if re_first_line:
                    top_line_match = re_first_line.match(line)
                    if top_line_match:
                        parse_header(top_log, context, top_line_match)
                        context.batch_count += 1
                        line_handled = True

                if not line_handled:
                    #  There won't be a current_entry if the start of the file
                    #  is corrupted or we're skipping content until
                    #  self.config.start_time
                    if context.current_entry:
                        # Expecting a process line at this point
                        # print(f"{context.line_count} : >{line}<", file=sys.stderr)
                        process_match = context.re_process.match(line)

                        if process_match:
                            if self.config.verbosity > 4:
                                print(
                                    f"process line match groups: {process_match.groups()}"
                                )
                            self._parse_process_line(
                                line, context.current_entry["timestamp"], process_match
                            )
                        else:
                            if self.config.verbosity > 3:
                                print(f"process line no match: '{line}'")

        if context.current_entry:  # stash the final entry
            self._top_entries.append(context.current_entry)

        if self._highest_cpuid is None:
            # Rather than expend effort on max() calls, assume that the header is sorted
            self._highest_cpuid = context.cpu_id

    # --------------------------------------------------------------------------

    def _parse_busybox_header(
        self, top_log: IOBase, context: ParseContext, header_match: re.Match
    ) -> None:
        """Parse a BusyBox format header block."""

        # When starting a new entry, stash previous one
        if context.current_entry:
            self._top_entries.append(context.current_entry)

        groupdict = header_match.groupdict()

        context.current_entry = groupdict

        # BusyBox doesn't emit a timestamp, so fake one with the batch count.
        # Sad times :(
        current_secs = context.batch_count
        context.current_entry["timestamp"] = context.batch_count

        if context.first_secs is None:
            context.first_secs = current_secs

        if self.config.start_time:
            if isinstance(self.config.start_time, str):
                self.config.start_time = int(self.config.start_time)

            if current_secs < self.config.start_time:
                context.current_entry = None
                return

        if self.config.stop_time:
            if isinstance(self.config.stop_time, str):
                if (
                    self.config.stop_time[0:1] == "+"
                    or self.config.stop_time[0:1] == "<"
                ):
                    self.config.stop_time = (
                        context.first_secs + self.config.stop_time[1:]
                    )

            if current_secs >= self.config.stop_time:
                context.current_entry = None
                return

        # BusyBox header lines are of structure:
        #
        #   Mem: 27436964K used, 5056536K free, 317416K shrd, 1433660K buff, 11810180K cached
        #   CPU:   5% usr   3% sys   0% nic  83% idle   7% io   0% irq   0% sirq
        #   Load average: 2.44 2.24 2.30 3/2473 953092
        #     PID  PPID USER     STAT   VSZ %VSZ %CPU COMMAND
        #    7418  7417 jonathan S     606m   2%   1% kitty
        #
        # But:
        #
        #  CPU may be CPU${n} with multiple lines
        #  Process lines may have an extra CPU column indicating cpu_id

        have_all_expected_header_lines = True
        pull_line = True
        line = ""
        for regex in [
            self._re_busybox_cpu,
            self._re_busybox_load_average,
        ]:
            if pull_line:
                line = top_log.readline()
            pull_line = True

            if line:
                context.line_count += 1
                match = regex.match(line)
                if match:
                    if regex is self._re_busybox_cpu:
                        context.cpu_id = match.group("cpu_id")
                        if context.cpu_id is not None:
                            # Handle split out CPU info
                            self.config.has_cpu_rows = True
                            pull_line = False
                            while match:
                                context.line_count += 1
                                context.cpu_id = match.group("cpu_id")
                                temp_dict = match.groupdict()
                                for key in [
                                    "user",
                                    "system",
                                    "nice",
                                    "idle",
                                    "wait",
                                    "hw_irq",
                                    "sw_irq",
                                ]:
                                    temp_dict[
                                        f"cpu{context.cpu_id}_{key}"
                                    ] = temp_dict.pop("cpu_" + key)
                                context.current_entry.update(temp_dict)

                                # Handle multiple cpu_id blocks on a single line
                                line = line[match.end() :]
                                line = line.lstrip()
                                if not line:
                                    line = top_log.readline()
                                match = regex.match(line)
                            context.line_count -= 1
                        else:
                            context.current_entry.update(match.groupdict())
                    else:
                        context.current_entry.update(match.groupdict())
                else:
                    have_all_expected_header_lines = False
                    print(
                        f">ERR: line {context.line_count}: >{line}<\n   :"
                        f" Unexpected match failure for {regex.pattern}",
                        file=sys.stderr,
                    )
            else:
                have_all_expected_header_lines = False

        if have_all_expected_header_lines:
            # Process header line
            line = top_log.readline()
            process_header_pattern = ""
            if not self._re_process_header:
                (
                    process_header_pattern,
                    self._re_process_header,
                    context.re_process,
                ) = self._generate_busybox_re_process(line)
                if re.search(r"[+*]CPU ", self._re_process_header.pattern):
                    self.config.has_cpu_column = True
            if not self._re_process_header.match(line):
                print(
                    f">ERR: line {context.line_count}: expected a process header"
                    f" line ({process_header_pattern}),\n got>{line}<",
                    file=sys.stderr,
                )

            context.line_count += 2

    # --------------------------------------------------------------------------

    def _parse_procps_header(
        self, top_log: IOBase, context: ParseContext, header_match: re.Match
    ) -> None:
        """Parse a procps format header block."""

        # starting a new entry, so stash previous one
        if context.current_entry:
            self._top_entries.append(context.current_entry)

        groupdict = header_match.groupdict()

        try:
            assert groupdict["timestamp"] is not None
        except AssertionError:
            print(f"ERR: Expected a timestamp field at line {context.line_count} !?")
            return

        context.current_entry = groupdict

        # Handle start/stop times
        current_secs = dhms_to_sec(context.current_entry["timestamp"])

        if context.first_secs is None:
            context.first_secs = current_secs

        if context.prev_secs is not None:
            # Handle midnight wrapping
            # Not sure that
            while current_secs < context.prev_secs:
                current_secs += 24 * 60 * 60

        # Convert to play nicely with matplotlib
        context.current_entry["timestamp"] = current_secs

        context.prev_secs = current_secs

        if self.config.start_time:
            if isinstance(self.config.start_time, str):
                offset = dhms_to_sec(self.config.start_time[1:])

                self.config.start_time = current_secs + offset

            if current_secs < self.config.start_time:
                context.current_entry = None
                return

        if self.config.stop_time:
            if isinstance(self.config.stop_time, str):
                offset = dhms_to_sec(self.config.stop_time[1:])

                if self.config.stop_time[0:1] == "+":
                    self.config.stop_time = current_secs + offset

                elif self.config.stop_time[0:1] == "<":
                    self.config.stop_time = context.first_secs + offset

            if current_secs >= self.config.stop_time:
                context.current_entry = None
                return

        # By default the header lines are of structure:
        #   Tasks:
        #   Cpu(s):
        #   Mem:
        #   Swap:
        #
        #   PID USER      PR  NI  VIRT  RES  SHR S %CPU %MEM    TIME+  COMMAND
        #
        # But Cpu can be aggregate or split out by core, and the process header
        # line is mutable

        have_all_expected_header_lines = True
        pull_line = True
        line = ""
        for regex in [
            self._re_procps_tasks,
            self._re_procps_cpu,
            self._re_procps_mem,
            self._re_procps_swap,
        ]:
            if pull_line:
                line = top_log.readline()
            pull_line = True

            if line:
                context.line_count += 1
                match = regex.match(line)
                if match:
                    if regex is self._re_procps_cpu:
                        context.cpu_id = match.group("cpu_id")
                        if context.cpu_id is not None:
                            # Handle split out CPU info
                            self.config.has_cpu_rows = True
                            pull_line = False
                            while match:
                                context.line_count += 1
                                context.cpu_id = match.group("cpu_id")
                                temp_dict = match.groupdict()
                                for key in [
                                    "user",
                                    "system",
                                    "nice",
                                    "idle",
                                    "wait",
                                    "hw_irq",
                                    "sw_irq",
                                    "steal",
                                ]:
                                    temp_dict[
                                        f"cpu{context.cpu_id}_{key}"
                                    ] = temp_dict.pop("cpu_" + key)
                                context.current_entry.update(temp_dict)

                                # Handle multiple cpu_id blocks on a single line
                                line = line[match.end() :]
                                line = line.lstrip()
                                if not line:
                                    line = top_log.readline()
                                match = regex.match(line)
                            context.line_count -= 1
                        else:
                            context.current_entry.update(match.groupdict())
                    else:
                        context.current_entry.update(match.groupdict())
                else:
                    have_all_expected_header_lines = False
                    print(
                        f">ERR: line {context.line_count}: >{line}<\n   :"
                        f" Unexpected match failure for {regex.pattern}",
                        file=sys.stderr,
                    )
            else:
                have_all_expected_header_lines = False

        if have_all_expected_header_lines:
            # Blank line
            line = top_log.readline()  # blank line
            if len(line) != 1:  # 1 for newline char
                print(
                    f">ERR: line {context.line_count}: >{line}<\nExpected a blank line"
                    " here.",
                    file=sys.stderr,
                )

            # Process header line
            line = top_log.readline()
            process_header_pattern = ""
            if not self._re_process_header:
                (
                    process_header_pattern,
                    self._re_process_header,
                    context.re_process,
                ) = self._generate_procps_re_process(line)
                if " +P " in self._re_process_header.pattern:
                    self.config.has_cpu_column = True
                    # TODO: confirm no longer required
                    # if not self.config.has_cpu_rows:
                    #    print(
                    #        "Currently when tracking per process CPU core"
                    #        " use, topplot needs per core CPU information"
                    #        " too."
                    #    )
                    #    print(
                    #        "For more information see the 'Workaround'"
                    #        " section"
                    #        " here:"
                    #        " https://gitlab.com/eBardie/topplot/-/issues/11"
                    #    )
                    #    sys.exit(1)
            if not self._re_process_header.match(line):
                print(
                    f">ERR: line {context.line_count}: expected a process header"
                    f" line ({process_header_pattern}),\n got>{line}<",
                    file=sys.stderr,
                )

            context.line_count += 2

    # --------------------------------------------------------------------------

    def _parse_process_line(self, line: str, timestamp: int, process_match: re.Match):
        """Parse a process line."""
        groupdict = process_match.groupdict()

        pid = process_match.group("pid")

        if self.config.toptype == TopType.BUSYBOX:
            # BusyBox *can* merge PID with PPID, so we alway parse PPID in to PID.
            # This may leave a space in there, so remove it now for consistency.
            pid = pid.replace(" ", "")

        del groupdict["pid"]

        command = process_match.group("command")
        del groupdict["command"]

        # Just the basename please
        if command[0:1] == "/":
            slash = command.rfind("/")
            if slash > -1:
                command = command[slash + 1 :]

        if command == "" or len(command) == 0:
            print(f"WARN: command empty for line: {line}")

        # Skip this process if it matches the ignore regex and doesn't match
        # the POI regex
        if (
            self._re_ignore_processes_regex
            and self._re_ignore_processes_regex.match(command)
        ) and not (self.config.poi_regex and self.config.poi_regex.match(command)):
            return

        if command not in self._processes:
            self._processes[command] = {}

        if pid not in self._processes[command]:
            self._processes[command][pid] = {}
            self._processes[command][pid]["timestamps"] = {}

            # Storing at this point obviously won't cope with processes that
            # update their own ARGV[0]
            self._processes[command][pid]["commandline"] = groupdict["commandline"]

            if self.config.poi_acc_cpu:
                self._processes[command][pid]["acc_cpu"] = 0
            if self.config.poi_acc_mem:
                self._processes[command][pid]["acc_mem"] = 0
            if self.config.poi_peak_cpu:
                self._processes[command][pid]["max_cpu"] = 0
            if self.config.poi_peak_mem:
                self._processes[command][pid]["max_mem"] = 0

        # Bizarrely procps top can throw out reports where all processes' CPU
        # column entries are negative
        cpu_value = float(groupdict["cpu"])
        if cpu_value < 0:
            groupdict["cpu"] = str(cpu_value * -1)

        if self.config.poi_acc_cpu:
            self._processes[command][pid]["acc_cpu"] += float(groupdict["cpu"])

        if self.config.poi_acc_mem:
            self._processes[command][pid]["acc_mem"] += float(groupdict["mem"])

        if self.config.poi_peak_cpu:
            cpu = float(groupdict["cpu"])
            if cpu > self._processes[command][pid]["max_cpu"]:
                self._processes[command][pid]["max_cpu"] = cpu

        if self.config.poi_peak_mem:
            mem = float(groupdict["mem"])
            if mem > self._processes[command][pid]["max_mem"]:
                self._processes[command][pid]["max_mem"] = mem

        self._processes[command][pid]["timestamps"][timestamp] = groupdict

        if self.config.toptype == TopType.BUSYBOX:
            # BusyBox batch mode (at least <= 1.35.1) has no way of getting
            # per core CPU info, so spend energy extracting max CPU core on
            # every line :(
            if "cpu_id" in groupdict:
                if self._highest_cpuid is None:
                    self._highest_cpuid = 0
                cpu_id = int(groupdict["cpu_id"])
                if cpu_id > self._highest_cpuid:
                    self._highest_cpuid = cpu_id

    # --------------------------------------------------------------------------



# Copyright (c) 2019-2020 Jonathan Sambrook and Codethink Ltd.
#
#    This file is part of Topplot.
#
#    Topplot is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Topplot is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Topplot.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------------
# Grapher has one or more FigManagers.
# FigManager has one figure with one or more plots/subplots and implements
# per-figure functionality.
# Grapher constructs figures, plots the data, and performs supra-FigManager
# functions too.
# -----------------------------------------------------------------------------

# pylint: disable=too-many-lines # C0302: Too many lines in module (>1000)

import sys
from typing import Callable, Dict

import matplotlib as mpl
import matplotlib.dates as mdates
import matplotlib.pyplot as plt

from ..config import Config
from ..dataframes import Dataframes
from ..figman import FigManager
from ..linestyler import LineStyler
from ..progress_window import ProgressWindow
from ..rcparams import override_rcParams
from ..graphmap import GraphMap
from ..utils import die, warn

try:
    import mplcursors

    mplcursors_present = True
except ImportError:
    print(
        "The mplcursors Python module is not installed, so annotations are not"
        " available. Hint: pip3 install mplcursors"
    )
    mplcursors_present = False

# -----------------------------------------------------------------------------
# Select GUI toolkit

mpl.use("TKAgg")

# -----------------------------------------------------------------------------


def add_annotations(figman, ax):
    """Add mplcursors annotation capability for the given axis' lines, if available."""
    if "mplcursors" in sys.modules:
        c = mplcursors.Cursor(hover=False, multiple=True, artists=tuple(ax.get_lines()))

        # At the time of writing (mpl 3.1.0) mpl doesn't allow consuming of gui
        # events, so they're propogated to all hooks. Which means that
        # mplcursors annotations are triggered even when they are underneath
        # legends.
        # This on_add closure callback fixes that situation.
        def on_add(sel):
            # Convert from data co-ords to display co-ords
            x, y = ax.transData.transform(tuple(sel.target))

            # Cycle through axes' legends checking for hits
            fake_event = mpl.backend_bases.MouseEvent("dummy", figman.fig.canvas, x, y)

            for legend in figman.get_legends():
                result, _ = legend.contains(fake_event)

                # Remove sel on hit
                if result:
                    c.remove_selection(sel)

        c.connect("add", on_add)


# -----------------------------------------------------------------------------


def set_x_axis_time_formatting(ax: mpl.axis.XAxis):
    """Turn on H:M:S date format for the given axis."""
    ax.xaxis.axis_date()
    ax.xaxis.set_major_formatter(mdates.DateFormatter("%H:%M:%S"))


# -----------------------------------------------------------------------------


class Grapher:
    """Class to encapsulate the graphing side of topplot."""

    # pylint: disable=import-outside-toplevel
    from ._cpus import graph_cpus, graph_cpu_per_cpu
    from ._legends import single_legend, combined_legend, otm_legend
    from ._mem import graph_mem
    from ._overview import graph_overview, graph_by_overview_ordinal
    from ._poi import graph_poi, graph_poi_per_cpu
    from ._tasks import graph_procps_tasks, graph_busybox_tasks

    # -------------------------------------------------- -----------------------

    def __init__(
        self,
        graph_map: GraphMap,
        config: Config,
        data: Dataframes,
        mem_unit: str,
        mem_cached_or_available: str,
        progress_window: ProgressWindow,
    ):
        self.at_close : Callable[[None], None]
        self.config = config
        self.data = data
        self.mem_unit = mem_unit
        self.graph_map = graph_map
        self.progress_window = progress_window
        width, height = progress_window.get_curr_screen_dimensions(scale=0.9)
        self.window_size = f"{width}x{height}"

        self.display_legends = True

        self.fig_manager: Dict[str, FigManager] = {}

        self.mem_cached_or_available = mem_cached_or_available.replace("_", " ")
        self.mplcursors_present = mplcursors_present

        self.linestyler = LineStyler()

        self.colours = {
            "combined": "red",
            "poi_cpu": "red",
            "poi_mem": "blue",
            "user": "green",
            "system": "red",
            "nice": "blue",
            "idle": (0.3, 0.3, 0.3),
            "wait": "black",
            "hw irq": "orange",
            "sw irq": "cyan",
            "steal": "gray",
            "exec": "chartreuse",
            "task running": "green",
            "task sleeping": "blue",
            "task stopped": "red",
            "task zombie": "black",
            "mem used": "blue",
            "mem free": "green",
            "mem buffers": "pink",
            self.mem_cached_or_available: "green",
            "swap free": "purple",
            "load average": "purple",
            "fig_face": (  # Equivalent to ProgressWindow's __init__().bg = "#F2F2E6".
                0.95,
                0.95,
                0.90,
            ),  # Keep in sync.
        }

        self.markers = {
            "user": "$u$",
            "system": "$s$",
            "nice": "$n$",
            "idle": "$i$",
            "wait": "$w$",
            "hw irq": "$hw$",
            "sw irq": "$sw$",
            "steal": "$st$",
            "exec": "$x$",
        }

        if self.config.cores > 1:
            for core in range(self.config.cores):
                tmp = {}
                for (k, v) in self.markers.items():
                    tmp[f"{core}{k}"] = f"{v}{core}"
                if tmp:
                    self.markers.update(tmp)

        self.cartesian_map = [["", ""], ["", ""]]

    # --------------------------------------------------------------------------

    def ordinal_to_title(self, n: int) -> str:
        """Convert from four-by-one array index to title stored in two-by-two array index."""
        assert (0 <= n <= 3), f"n must be between 0 and 3 but is {n}"
        x = int(n / 2)
        y = n % 2
        return self.cartesian_map[x][y]

    # --------------------------------------------------------------------------

    def title(self, text: str) -> str:
        """Common title formatting."""
        return f"topplot : {self.config.toplog_filename} : {text}"

    # --------------------------------------------------------------------------

    @staticmethod
    def common_axes(figman: FigManager, ax: mpl.axis.Axis):
        """Styles for any/all axes."""
        ax.set_facecolor("white")
        ax.margins(0)
        ax.set_xlabel("")
        set_x_axis_time_formatting(ax)
        add_annotations(figman, ax)

    # --------------------------------------------------------------------------

    def common_ax1(self, figman: FigManager, ax: mpl.axis.Axis):
        """Styles for 'primary' axes."""
        self.common_axes(figman, ax)
        ax.grid(linestyle=":", linewidth="0.5", color="black", alpha=0.5)

    # --------------------------------------------------------------------------

    def common_ax2(self, figman: FigManager, ax: mpl.axis.Axis):
        """Styles for 'secondary' axes."""
        self.common_axes(figman, ax)
        ax.grid(linestyle="--", linewidth="0.5", color="black", alpha=0.75)

    # --------------------------------------------------------------------------

    def dispense_graph(
        self,
        title: str,
        figman_name: str,
        setup: Callable[[FigManager], None],
        generate_figman: Callable[[None], None],
        use_progress_window: bool,
    ):
        """Handle progress window when creating graph."""
        if figman_name in self.fig_manager:
            self.fig_manager[figman_name].show()
        else:

            def create_graph(self):
                figman = generate_figman()

                figman.setup_graphing(setup)

                if use_progress_window:
                    self.progress_window.hide()

                self.fig_manager[figman_name] = figman

            if use_progress_window:
                self.progress_window.update_status(f"preparing '{title}'")
                self.progress_window.show()
                self.progress_window.root.after(100, create_graph, self)
            else:
                create_graph(self)

    # --------------------------------------------------------------------------

    def figs_foo(
        self, func: Callable[[FigManager], None], all_figs: bool, name: str = None
    ):
        """Call specified method on all the figs, the specified one,
        or the most recently created one."""
        if self.fig_manager:
            figs = list(self.fig_manager.values())
            if name is not None:
                figs = next(filter(lambda x: x == name, figs))

            if figs:
                to_do_list = figs if all_figs else [figs[-1]]
                for figman in to_do_list:
                    if figman:
                        func(figman)

    # --------------------------------------------------------------------------

    def save_figs(self, all_figs: bool = True, name: str = None):
        """Save pngs of all the figs, the specified one, or the most recently created one.."""
        self.figs_foo(FigManager.save, all_figs, name)

    # --------------------------------------------------------------------------

    def close_figs(self, all_figs: bool = True, name: str = None):
        """Close all the figs, the specified one, or the most recently created one.."""
        self.figs_foo(FigManager.close, all_figs, name)
        self.close_check()

    # --------------------------------------------------------------------------

    def close_check(self):
        """FigManagers call this when closing to exit the app if they were the last one."""
        if not self.fig_manager and self.at_close is not None:
            self.at_close()

    # --------------------------------------------------------------------------

    def prep_cartesian_title_map(self):
        for (x, y, title) in zip([0, 0, 1, 1], [0, 1, 0, 1], self.graph_map.keys()):
            self.cartesian_map[x][y] = title

    # --------------------------------------------------------------------------

    def doit(
        self, at_showtime: Callable[[None], None], at_close: Callable[[None], None]
    ):
        """Main function for setting things up and running."""
        self.at_close = at_close

        override_rcParams()

        self.prep_cartesian_title_map()

        count = 1

        if len(self.config.which_graphs) > 0:
            count = 0
            for c in self.config.which_graphs:
                if c.isnumeric():
                    i = int(c)
                    if i == 0:
                        self.graph_overview()
                        count += 1

                    elif 1 <= i <= 4:
                        self.graph_by_overview_ordinal(i - 1)
                        count += 1

                    else:
                        warn("'-g' takes numbers from 0 to 4.")

                elif c == "c":
                    if self.config.has_cpu_column and self.config.cores > 1:
                        self.graph_poi_per_cpu()
                        count += 1
                    else:
                        warn("No multi-core data available to plot POI by CPU core.")

                elif c == "C":
                    if self.config.has_cpu_rows and self.config.cores > 1:
                        self.graph_cpu_per_cpu()
                        count += 1
                    else:
                        warn("No multi-core data available to plot CPU data by core.")

                elif c == "p":
                    if count > 0:
                        self.save_figs(all_figs=False)

                elif c == "P":
                    if count > 0:
                        self.save_figs()

                elif c == "q":
                    if count > 0:
                        self.close_figs(all_figs=False)
                        if count > 1:
                            count -= 1
                        else:
                            break
                    else:
                        break

                elif c == "Q":
                    if count > 0:
                        self.close_figs()
                        count = 0
                    break

                else:
                    die(f"The character '{c}' doesn't represent a graph.")

        else:
            self.graph_overview()

        at_showtime()
        if count > 0:
            plt.show()


# -----------------------------------------------------------------------------
# vi: sw=4:ts=4:et



# -------------------------------------------------------------------------


def graph_procps_tasks(
    self,
    figman,
    df_in,
    ax_sleeping,
    title,
    overview: bool = False,  # pylint: disable=unused-argument
):
    """Draw the procps task summary graph."""

    ax_others = ax_sleeping.twinx()

    df = df_in.rename(columns=lambda x: x[5:] if x[:5] == "task " else x)
    df.filter(items=["sleeping"]).plot(
        ax=ax_sleeping, color=self.colours["task sleeping"], x_compat=True
    )

    task_colours = []
    items = ["running", "stopped", "zombie"]
    for item in items:
        task_colours.append(self.colours["task " + item])
    df.filter(items=items).plot(ax=ax_others, color=task_colours, x_compat=True)

    self.combined_legend(
        figman,
        title,
        "tasks",
        "sleeping tasks",
        "running, stopped, and zombie tasks",
        ax_sleeping,
        ax_others,
    )
    self.common_ax1(figman, ax_sleeping)
    self.common_ax2(figman, ax_others)
    ax_others.set_ybound(lower=0)
    ax_sleeping.set_ybound(lower=0, upper=df["sleeping"].max() * 105.0 / 100)

# -------------------------------------------------------------------------


def graph_busybox_tasks(
    self,
    figman,
    df_in,
    ax_in,
    title,
    overview: bool = False,  # pylint: disable=unused-argument
):
    """Draw the BusyBox task summary graph."""
    ax_in.set_title(title)
    df = df_in.rename(columns=lambda x: x[5:] if x[:5] == "task " else x)
    df.plot(ax=ax_in, color=self.colours["task running"], x_compat=True)
    self.single_legend(figman, title, "tasks", "count of running tasks", ax_in)
    self.common_ax1(figman, ax_in)

# -------------------------------------------------------------------------



import math
from typing import Any, Dict, List, Tuple
from ..figman import FigManager

# -------------------------------------------------------------------------


def generate_markeveries(
    row_count: int, col_count: int, total_markers: int
) -> List[Tuple[int, int]]:
    """Generate a list of (offset, markevery) tuples for a dataset of the given dimensions.

    Makes an attempt to avoid clustering markers on different lines.

    Args:
        row_count (int): How many rows in the dataset.
        col_count (int): How many columns in the dataset.
        total_markers (int): How many markers are required for each line.

    Returns:
        list: The generated (offset, markevery) tuples.
    """
    markeveries = []

    # Calculate offsets for markers so that they don't cluster
    clustering = True

    markevery = int(row_count / total_markers)

    # Loop if offsets are (still) clustered and another attempt makes sense
    while clustering and total_markers >= 1:
        offset_prev = None
        for col in range(col_count):
            offset = int(markevery / col_count) * col
            if offset_prev is not None and offset_prev != offset:
                clustering = False
            offset_prev = offset
            markeveries.append((offset, markevery))
        total_markers -= 1

    return markeveries

# -------------------------------------------------------------------------


def graph_cpus(
    self,
    figman,
    df,
    ax_in,
    title,
    *,
    x_bounds=None,
    force_single_core: bool = None,
    overview: bool = False,
):
    """Draw the load average and CPU graph.

    Forcing a single cpu core isn't the same as there being no multicore data.

    Forcing a single core displays the core's index in labels, and removes the
    load average display since that's not relevant to a single core out of
    multiple cores.
    """
    ax_loadavg = None

    if force_single_core is None:
        ax_loadavg = ax_in
        ax_cpu = ax_loadavg.twinx()
        legend_title = "load and cpu"
    else:
        ax_cpu = ax_in
        legend_title = "cpu"

    if force_single_core is None:
        df.plot(
            y="load average",
            color=self.colours["load average"],
            ax=ax_loadavg,
            lw=3,
        )

    measures = ["user", "system", "nice", "idle", "wait", "hw irq", "sw irq"]

    if self.config.with_cpu_steal:
        measures.append("steal")

    core_test = 0 if force_single_core is None else force_single_core
    colname_test = f"cpu{core_test} exec" if self.config.has_cpu_rows else "cpu exec"
    if colname_test in df.columns:
        measures = ["exec"] + measures

    if self.config.has_cpu_rows or force_single_core is not None:
        # never display line markers for overview, and only if ordered to otherwise
        display_cpu_markers = False if overview else self.config.display_cpu_markers

        if display_cpu_markers:
            # To put text markers on each line has two cost implications
            #  i) df.plot() doesn't handle lists of markers/markeveries, so
            #     you need a two level loop and (cores * len(measures)) calls
            #     to df.plot()
            # ii) the markers trigger mathstext font related calls.
            #     Informative result, but slow.

            markeveries = self.generate_markeveries(len(df), self.config.cores, 10)

            for core in (
                range(0, self.config.cores)
                if force_single_core is None
                else [force_single_core]
            ):
                for measure in measures:
                    extra_args : Dict[str, Any] = {}
                    if force_single_core is None:
                        # The dollar symbols trigger lots of calling of
                        # expensive mathtext font functions
                        extra_args["marker"] = f"${core}$"
                        extra_args["markevery"] = markeveries[core]
                        extra_args["markersize"] = 8

                    df.plot(
                        y=f"cpu{core} {measure}",
                        color=self.colours[measure],
                        ax=ax_cpu,
                        **extra_args,
                    )
        else:  # This is the fast-but-no-markers version
            labels = []
            colours = []

            for measure in measures:
                for core in (
                    range(0, self.config.cores)
                    if force_single_core is None
                    else [force_single_core]
                ):
                    labels.append(f"cpu{core} {measure}")
                    colours.append(self.colours[measure])

            df.plot(
                y=labels,
                color=colours,
                ax=ax_cpu,
            )
    else:  # Single core (forced or natural)
        labels = []
        colours = []

        for measure in measures:
            labels.append(f"cpu {measure}")
            colours.append(self.colours[measure])

        df.plot(
            y=labels,
            color=colours,
            ax=ax_cpu,
        )

    # Legend(s)
    if ax_loadavg and ax_cpu:
        combined_legend = self.combined_legend(
            figman,
            title,
            legend_title,
            "loadavg",
            "cpu (%)",
            ax_loadavg,
            ax_cpu,
            location="upper left",
        )  # , cap2=len(measures))

        self.otm_legend(
            figman,
            "loadavg and\ngrouped cpu" if self.config.has_cpu_rows else "loadavg and cpu",
            # strip any leading "cpu0 ", "cpu1 ", ..
            r"^(?:cpu\d+ )?(.+)",
            ax_cpu,
            combined_legend,
            copy_markers=False,
        )

        if not (force_single_core is None and self.config.cores > 1) or overview:
            combined_legend.remove()
    else:
        self.single_legend(figman, title, legend_title, "cpu (%)", ax_cpu)

    # Axes fettling
    if force_single_core is None and ax_loadavg:
        ax_loadavg.tick_params("x", which="minor", bottom=False)
        self.common_ax1(figman, ax_loadavg)
        ax_loadavg.set_ybound(lower=0, upper=df["load average"].max() * 105.0 / 100)

    if x_bounds is not None:
        min_timestamp, max_timestamp = x_bounds
        ax_cpu.set_xbound(lower=min_timestamp, upper=max_timestamp)

    self.common_ax2(figman, ax_cpu)
    ax_cpu.set_ybound(lower=0, upper=100)

# -------------------------------------------------------------------------


def graph_cpu_per_cpu(self, use_progress_window: bool = False):
    """Draw the cpu data in separate plots for each CPU core."""
    figman_name = "cpu_by_cpu"
    title = "cpu data by cpu core"

    # Find the nearest integer square that will fit the plots
    sqrt = math.ceil(math.sqrt(int(self.config.cores)))

    # Account for this possibly being a row larger than needed
    excess_rows = 0
    if (sqrt * sqrt) > (self.config.cores + sqrt - 1):
        excess_rows = 1

    def setup(figman):
        min_timestamp = self.data.poi_df.head(1).index[0]
        max_timestamp = self.data.poi_df.tail(1).index[0]

        for core in range(self.config.cores):
            if self.config.cores < 3:
                ax = figman.plots[core]
            else:
                x = int(core / sqrt)
                y = core % sqrt

                ax = figman.plots[x][y]
            df = self.data.cpus_df.filter(regex=f"^cpu{core}")

            self.graph_cpus(
                figman,
                df,
                ax,
                f"core {core}",
                x_bounds=(min_timestamp, max_timestamp),
                force_single_core=core,
            )

    def generate_figman():
        return FigManager(
            figman_name,
            self,
            self.title(title),
            self.window_size,
            sqrt - excess_rows,
            sqrt,
            share_legends_regex=r"^cpu\d+ (.*)$"
        )

    self.dispense_graph(
        title.partition("\n")[0],
        figman_name,
        setup,
        generate_figman,
        use_progress_window,
    )

# -------------------------------------------------------------------------



import re

import matplotlib as mpl
import matplotlib.pyplot as plt

# -------------------------------------------------------------------------


def pickable_legend_lines(figman, legend, ax1, ax2=None, len1=None):
    """Map legend lines to plotted lines.

       Note: An mpl API workaround is required.
    """
    leglines = legend.get_lines()
    legtexts = legend.get_texts()

    # WARNING: going off mpl API here to get the legend's handles
    leghandles = legend.legendHandles

    # WARNING: going off mpl API here to access the artist of the legend's title box
    title = legend._legend_title_box._text  # pylint: disable=protected-access

    if len1 is None:
        len1 = len(ax1.get_lines())

    if len1 > 0:
        figman.map_legend_lines(
            ax1,
            title,
            leglines=leglines[:len1],
            legtexts=legtexts[:len1],
            leghandles=leghandles[:len1],
        )

    if ax2 is not None:
        figman.map_legend_lines(
            ax2,
            title,
            leglines=leglines[len1:],
            legtexts=legtexts[len1:],
            leghandles=leghandles[len1:],
        )

# -------------------------------------------------------------------------


def single_legend(_, figman, title, legtitle, ax_ylabel, ax, *, cap=None):
    """Produce a legend for a single axis."""
    # Set furniture text
    ax.set_title(title)
    ax.set_ylabel(ax_ylabel)

    # Get legend lines ready
    handles, labels = ax.get_legend_handles_labels()

    if cap is not None:
        handles = handles[0:cap]
        labels = labels[0:cap]

    # Generate new legend
    legend = ax.legend(handles, labels, loc="upper right", title=legtitle)
    legend.set_draggable(True)

    pickable_legend_lines(figman, legend, ax)

# -------------------------------------------------------------------------


def otm_legend(
    _,
    figman,
    legtitle,
    prefix_regex,
    ax,
    src_legend,
    *,
    loc="upper right",
    copy_markers=True,
):
    """Produce an extra one-to-many (otm) legend to enable togging of a particular
       measurement in a single plot displaying all CPU cores.
    """
    re_prefix = re.compile(prefix_regex)
    re_trailing_digits = re.compile(r"\d+$")

    categories = {}

    # Sort out toggleable elements
    for (legline, handle, text) in zip(
        src_legend.get_lines(), src_legend.legendHandles, src_legend.texts
    ):
        label = text.get_text()
        match = re_prefix.match(label)
        if match is not None:
            category = match.group(1)
            legelts = [figman.legtexts[legline], figman.legmarkers[legline]]
            if category not in categories:
                marker = (
                    handle._marker.get_marker()  # pylint: disable=protected-access
                )
                marker = re_trailing_digits.sub("", marker, 1)
                categories[category] = {
                    "lines": [handle],
                    "legelts": legelts,
                    "colour": handle.get_color(),
                    "lw": handle.get_linewidth(),
                    "marker": marker,
                }
            else:
                categories[category]["lines"].append(handle)
                categories[category]["legelts"] += legelts

    # Create shiney new legend lines
    new_handles = []
    new_labels = []
    for category, cat_data in categories.items():
        extra_args = {"lw": cat_data["lw"]}
        if copy_markers:
            extra_args["marker"] = cat_data["marker"]
        line = mpl.lines.Line2D(
            [],
            [],
            color=cat_data["colour"],
            label=category,
            **extra_args,
        )
        new_handles.append(line)
        new_labels.append(category)

    # Generate new legend
    legend = plt.legend(new_handles, new_labels, loc=loc, title=legtitle)
    figman.register_legend(ax, legtitle, legend)
    legend.set_draggable(True)

    # WARNING: going off mpl API here to access the artist of the legend's title box
    title = legend._legend_title_box._text  # pylint: disable=protected-access
    title.set_picker(10)
    figman.legotmtitles[title] = []

    for (line, text) in zip(legend.get_lines(), legend.get_texts()):
        label = line.get_label()
        line.set_picker(5)
        text.set_picker(10)
        figman.legotm[line] = categories[label]["lines"]
        figman.legotm[text] = categories[label]["lines"]
        figman.legotmtitles[title].append(line)
        figman.legotmtexts[line] = text
        figman.legotmtexts[text] = line

    # Re-add src legend since it will have been detached by the
    # plt.legend([..]) call above
    ax.add_artist(src_legend)

    return legend

# -------------------------------------------------------------------------


def combined_legend(
    _,
    figman,
    title,
    legtitle,
    ax1_ylabel,
    ax2_ylabel,
    ax1,
    ax2,
    *,
    cap1=None,
    cap2=None,
    location="upper right",
):
    """Combine legends from both axes in to single axis."""
    # Set furniture text
    ax2.set_title(title)
    ax1.set_ylabel(ax1_ylabel)
    ax2.set_ylabel(ax2_ylabel)

    # Get legend lines ready
    handles1, labels1 = ax1.get_legend_handles_labels()
    handles2, labels2 = ax2.get_legend_handles_labels()

    if cap1 is not None:
        handles1 = handles1[0:cap1]
        labels1 = labels1[0:cap1]

    if cap2 is not None:
        handles2 = handles2[0:cap2]
        labels2 = labels2[0:cap2]

    # Remove original legends from drawing tree
    ax1.get_legend().remove()
    ax2.get_legend().remove()

    # Generate new legend
    legend = ax2.legend(
        handles1 + handles2, labels1 + labels2, loc=location, title=legtitle
    )
    legend.set_draggable(True)
    ax2.get_legend().set_draggable(True)

    pickable_legend_lines(figman, legend, ax1, ax2, len1=len(handles1))

    return legend

# -------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Early non-graphing output


def list_processes(self):
    for command in sorted(self.processes.keys()):
        pids = []
        for pid in sorted(self.processes[command].keys()):
            if self.config.list_processes >= 2:
                print(
                    f"{command} [{pid}]"
                    f" {self.processes[command][pid]['commandline']}"
                )
            else:
                pids.append(pid)

        if self.config.list_processes == 0:
            print(f"{command}")
        elif self.config.list_processes == 1:
            print(f"{command} x{len(pids)} {pids}")


# ------------------------------------------------------------------------------



import math

import matplotlib.pyplot as plt

from ..figman import FigManager

# -------------------------------------------------------------------------


def style_lines(self, *axes):
    """Apply the relevant style to the lines on the given axis/axes."""
    for ax in axes:
        prefix = ""
        label = ax.get_yaxis().get_label().get_text()
        if label is not None:
            prefix = label
        for line in ax.get_lines():
            self.linestyler.style(prefix + line.get_label(), line)

# -------------------------------------------------------------------------


def graph_poi(
    self,
    figman,
    df,
    ax,
    title,
    *,
    x_bounds=None,
    mem_bounds=None,
    single_core=None,
    overview: bool = False,  # pylint: disable=unused-argument
):
    """Draw the graph of Processes of Interest (POI) using the given DataFrame."""
    max_cpu = 100
    max_mem = None

    # Convenience aliasing
    plot_cpu_lines = self.config.plot_poi_cpu_lines
    plot_cpu_sum = self.config.plot_poi_cpu_sum
    plot_mem_lines = self.config.plot_poi_mem_lines
    plot_mem_sum = self.config.plot_poi_mem_sum

    if (plot_mem_lines or plot_mem_sum) and self.config.include_process_mem:
        ax_cpu = ax.twinx()
        ax_mem = ax
    else:
        ax_cpu = ax
        ax_mem = ax.twinx()

    figman.ax_name_map[ax_cpu] = "ax_cpu"
    figman.ax_name_map[ax_mem] = "ax_mem"

    if (
        (plot_cpu_lines or plot_cpu_sum)
        and self.config.include_process_cpu
        and (plot_mem_lines or plot_cpu_sum)
        and self.config.include_process_mem
    ):
        figman.ax_pairs.append(ax_cpu, ax_mem)

    # ---------------------------------------------------------------------

    def style_line(
        ax, index, colour, alpha, linestyle, marker=None, mark_every=None
    ):
        line = ax.get_lines()[index]
        line.set_color(colour)
        line.set_alpha(alpha)
        line.set_linestyle(linestyle)
        if marker is None:
            marker = ""
        line.set_marker(marker)
        if mark_every is None:
            mark_every = 10
        line.set_markevery(mark_every)

    # ---------------------------------------------------------------------

    def max_cpu_fu(column_name):
        max_cpu = df[column_name].max()
        if max_cpu > 100:
            max_cpu = 100 * math.ceil((max_cpu + 0.1) / 100.0)
        else:
            max_cpu = 110
        return max_cpu

    # ---------------------------------------------------------------------

    def max_mem_fu(column_name):
        max_mem = df[column_name].max()
        return 10 * math.ceil((max_mem + 0.1) / 10.0)

    # ---------------------------------------------------------------------

    def handle_data(
        df,
        ax,
        separate_cores,
        plot_data_lines,
        plot_summary,
        plot_at_all,
        mode,
        summary_line_style,
        max_fu,
        max_fu_default,
        x_bounds,
    ):
        plotted = False
        if (plot_data_lines or plot_summary) and plot_at_all:
            summary_title = f"poi {mode} sum - {mode}"
            if plot_summary and not separate_cores:
                # Create summary column if not already present
                if summary_title not in df.columns.to_list():
                    summary = self.data.poi_df.filter(regex=f" - {mode}$").sum(
                        axis=1
                    )
                    df.insert(0, summary_title, summary)

            # Select columns by regex
            if not plot_data_lines:
                regex = f"^{summary_title}$"
            else:
                regex = f"- {mode}$"

            # Strip trailing category from names then plot
            def stripper(x):
                return x[:-6] if x[-6:] == f" - {mode}" else x

            df = df.filter(regex=regex)
            if not df.empty:
                df.rename(columns=stripper).plot(ax=ax, xlim=x_bounds)
                plotted = True
            ax.set_ylabel(f"{mode} (%)")
            style_lines(self, ax)

            # Override summary line
            if plot_summary and not separate_cores:
                colour, alpha, linestyle, marker, mark_every = summary_line_style
                style_line(ax, 0, colour, alpha, linestyle, marker, mark_every)
                return (plotted, max_fu(summary_title))

        return (plotted, max_fu_default)

    # ---------------------------------------------------------------------

    total_markers = 7.5
    mark_every = int(len(df) / total_markers)
    alpha = 0.33
    cpu_summary_line_style = (
        "mediumvioletred",
        alpha,
        (1, (1, 2, 1, 3)),
        "$c$",
        mark_every,
    )
    mem_summary_line_style = (
        "dodgerblue",
        alpha,
        (0, (2, 6, 2, 2)),
        "$m$",
        mark_every,
    )

    cpu_plotted, max_cpu = handle_data(
        df,
        ax_cpu,
        x_bounds is not None,
        plot_cpu_lines,
        plot_cpu_sum,
        self.config.include_process_cpu,
        "cpu",
        cpu_summary_line_style,
        max_cpu_fu,
        100 if single_core else 110,
        x_bounds,
    )

    mem_plotted, max_mem = handle_data(  # pylint: disable=unused-variable
        df,
        ax_mem,
        x_bounds is not None,
        plot_mem_lines,
        plot_mem_sum,
        self.config.include_process_mem,
        "mem",
        mem_summary_line_style,
        max_mem_fu,
        None,
        x_bounds,
    )

    # Handle legends
    l_mem_cuckoo = None

    # Handle mem
    if (plot_mem_lines or plot_mem_sum) and self.config.include_process_mem:
        # ax_mem's legend shouldn't be overwritten by lines on ax_cpu
        # Sadly they are being overwritten, so ensure they're not by adding
        # them to ax_cpu
        if (
            (plot_cpu_lines or plot_cpu_sum)
            and self.config.include_process_cpu
            and cpu_plotted
        ):
            handles, labels = ax_mem.get_legend_handles_labels()
            # Remove old legend's artists from drawing tree to avoid
            # bitching about artist reuse
            # Note: doesn't delete legend object
            old_l_mem = ax_mem.legend()
            old_l_mem.remove()
            l_mem_cuckoo = plt.legend(
                handles, labels, loc="upper left", title="mem"
            )

        else:
            ax_mem.legend(loc="upper left", title="mem")
            ax_cpu.set_visible(False)

            l_mem = ax_mem.get_legend()
            # WARNING: accessing internal mpl details here
            label = (
                l_mem._legend_title_box._text  # pylint: disable=protected-access
            )
            figman.map_legend_lines(ax_mem, label, l_mem)
            l_mem.set_draggable(True)
            ax_mem.xaxis.set_visible(True)
            ax_mem.patch.set_visible(True)

    # Handle cpu
    if (plot_cpu_lines or plot_cpu_sum) and self.config.include_process_cpu:
        l_cpu = ax_cpu.legend(loc="upper right", title="cpu")
        # WARNING: accessing internal mpl details here
        label = l_cpu._legend_title_box._text  # pylint: disable=protected-access
        figman.map_legend_lines(ax_cpu, label, l_cpu)
        l_cpu.set_draggable(True)

        if (
            not plot_mem_lines and not plot_mem_sum
        ) or not self.config.include_process_mem:
            ax_mem.set_visible(False)

    if l_mem_cuckoo is not None:
        ax_cpu.add_artist(l_mem_cuckoo)
        figman.register_legend(ax_cpu, "cuckoo", l_mem_cuckoo)
        # WARNING: accessing internal mpl details here
        label = (
            l_mem_cuckoo._legend_title_box._text  # pylint: disable=protected-access
        )
        figman.map_legend_lines(ax_mem, label, l_mem_cuckoo)
        l_mem_cuckoo.set_draggable(True)

    # Handle furniture
    ax_cpu.set_title(title)
    ax_mem.set_title(title)

    self.common_ax1(figman, ax_mem)
    self.common_ax2(figman, ax_cpu)

    if x_bounds is None:
        x_bounds = (df.head(1).index[0], df.tail(1).index[0])
    ax_cpu.set_xlim(x_bounds)

    ax_cpu.set_ybound(lower=0, upper=max_cpu)

    if mem_bounds is not None:
        min_mem, max_mem = mem_bounds
        ax_mem.set_ybound(lower=min_mem, upper=max_mem)
    elif max_mem is not None:
        ax_mem.set_ybound(lower=0, upper=max_mem)

# -------------------------------------------------------------------------


def graph_poi_per_cpu(self, use_progress_window: bool = False):
    """Draw the Processes of Interest in separate plots for each CPU core."""
    figman_name = "poi_by_cpu"
    title = "poi by cpu core"

    # Find the nearest integer square that will fit the plots
    sqrt = math.ceil(math.sqrt(int(self.config.cores)))

    # Account for this possibly being a row larger than needed
    excess_rows = 0
    if (sqrt * sqrt) > (self.config.cores + sqrt - 1):
        excess_rows = 1

    def setup(figman):
        min_timestamp = self.data.poi_df.head(1).index[0]
        max_timestamp = self.data.poi_df.tail(1).index[0]

        max_mem = 0
        for core in range(self.config.cores):
            max_mem = max(
                max_mem,
                self.data.core_dfs[core].filter(regex=" - mem$").max().max(),
            )

        for core in range(self.config.cores):
            if self.config.cores < 3:
                ax = figman.plots[core]
            else:
                x = int(core / sqrt)
                y = core % sqrt
                ax = figman.plots[x][y]
            self.graph_poi(
                figman,
                self.data.core_dfs[core],
                ax,
                f"core {core}",
                x_bounds=(min_timestamp, max_timestamp),
                mem_bounds=(0, max_mem),
                single_core=core,
            )

    def generate_figman():
        return FigManager(
            figman_name,
            self,
            self.title(f"{title}\n{self.config.poi_categories}"),
            self.window_size,
            sqrt - excess_rows,
            sqrt,
            subtitle=True,
            share_legends_regex="^(.*)$"
        )

    self.dispense_graph(
        title.partition("\n")[0],
        figman_name,
        setup,
        generate_figman,
        use_progress_window,
    )

# -------------------------------------------------------------------------



from .gui_utils import *
from .utils import *

from .config import Config

from .linestyler import LineStyler
from .figman import FigManager
from .rcparams import rcParams

from .graphs import Grapher
from .parser import Parser

from .app import App



# ------------------------------------------------------------------------------
# Since top has variance in output across versions, handle regex variants


class Re_Variants:
    def __init__(self, name, first_entry, *, var_substs=None, scope=None):
        self._name = name
        self.index = 0
        self.re_arr = [first_entry]
        assert (
            var_substs is None or scope is not None
        ), "Re_Variants: var_substs given without corresponding scope"
        self.var_substs = [var_substs]
        self.scope = scope

    # --------------------------------------------------------------------------

    def append(self, re, var_substs=None, scope=None):
        self.re_arr.append(re)
        if var_substs is not None:
            assert (
                scope is not None or self.scope is not None
            ), "Re_Variants: var_substs given without corresponding scope"
            self.var_substs.append(var_substs)
            if scope:
                self.scope = scope

    # --------------------------------------------------------------------------

    def name(self):
        return self._name

    # --------------------------------------------------------------------------
    # Override match([..]) to fallback on regex variants
    def match(self, *args, **kwargs):
        i = self.index
        response = None

        while response is None and i < len(self.re_arr):
            response = self.re_arr[i].match(*args, **kwargs)
            if response is None:
                i += 1
            else:
                if i != self.index:
                    if self.var_substs[i] is not None:
                        for (var_name, var_value) in self.var_substs[i]:
                            # pylint: disable=exec-used
                            exec(f'{var_name}="{var_value}"', self.scope)

        if response is not None:
            self.index = i

        return response

    # --------------------------------------------------------------------------

    def pattern(self):
        return [regex.pattern for regex in self.re_arr]

    # --------------------------------------------------------------------------
    # Delegate all other functionality to current regex
    def __getattr__(self, attr):
        print(f"attr: >{attr}<")
        return getattr(self.re_arr[self.index], attr)


# ------------------------------------------------------------------------------



from ..toptype import TopType

# -------------------------------------------------------------------------


def graph_mem(
    self,
    figman,
    df,
    ax1,
    title,
    overview: bool = False,  # pylint: disable=unused-argument
):
    """Draw the memory summary graph."""
    ax1.set_title(title)

    mem_colours = []

    # Draw 'mem free' lowest in z-order, unless 'mem availble' is needed there
    lowest_mem = (
        "mem free"
        if self.mem_cached_or_available == "mem cached"
        else "mem available"
    )

    items = [lowest_mem, "mem used", "mem buffers"]
    for item in items:
        mem_colours.append(self.colours[item])
    df.filter(items=items).plot.area(ax=ax1, color=mem_colours)

    # Inform the user if it looks like top's scale was set too high
    if df.iloc[[0, -1]].sum(axis=1).sum() == 0.0:
        msg = (
            "The memory and swap values for the first and last cycles are all zero."
            " Was top's scale set too high?\nWhen configuring top, toggle through"
            " the available scales by pressing uppercase 'E' in top's Interactive"
            " mode's main page.\nThen save the config by press uppercase 'W'."
        )
        figman.display_msg(msg, 5)
        print(f"WARNING: {msg}")

    # 'mem used' includes 'mem cached' which is pants, since that memory is
    # available to use immediately, although if the VFS is asked for its old
    # contents and they're still valid, it will (immediately) be put back to use
    # as that data instead.  ('mem available' handles this better.)
    #
    # Indicate this state of affairs by colouring cached memory using the colours
    # for 'mem free' hatched with 'mem used'

    if self.mem_cached_or_available == "mem cached":
        # Overwrite the overlapping part of 'mem used' with 'mem cached'
        df.filter(items=["mem free", "mem cached"]).plot.area(
            ax=ax1, color=self.colours["mem free"]
        )
        chatch = ax1.collections[-1]

        # Knock out confusing repeat 'mem free' legend
        c = ax1.collections[-2]
        c.set_label("")
    else:
        # Overwrite the overlapping part of 'mem available' with 'mem free'
        df.filter(items=["mem free"]).plot.area(
            ax=ax1, color=self.colours["mem free"]
        )
        chatch = ax1.collections[0]
        chatch.set_label("mem cached")

    chatch.set_facecolor(self.colours["mem free"])
    chatch.set_edgecolor(self.colours["mem used"])
    # hatches = [".", "/", "\\", None, "\\\\", "*"]
    chatch.set_hatch("//")

    if self.config.toptype == TopType.PROCPS:
        df.filter(items=["swap free"]).plot(
            ax=ax1, color=self.colours["swap free"], lw=3
        )

    self.common_ax1(figman, ax1)

    l1 = ax1.legend(loc="upper right", title="memory")
    l1.set_draggable(True)
    ax1.set_ylabel(f"mem ({self.mem_unit})")

# -------------------------------------------------------------------------



# ------------------------------------------------------------------------------


def get_display_offset(x, y):
    try:
        import screeninfo  # pylint: disable=import-outside-toplevel
    except ImportError:
        return (0, 0)

    def get_monitor_from_coord(x, y):
        monitors = screeninfo.get_monitors()

        for m in reversed(monitors):
            if m.x <= x <= m.width + m.x and m.y <= y <= m.height + m.y:
                return m
        return monitors[0]

    current_screen = get_monitor_from_coord(x, y)

    return (current_screen.x, current_screen.y)


# ------------------------------------------------------------------------------


def center_window_geometry(
    width, height, screen_width, screen_height, mouse_x, mouse_y
):
    offset_x, offset_y = get_display_offset(mouse_x, mouse_y)
    x = offset_x + (screen_width // 2) - (width // 2)
    y = offset_y + (screen_height // 2) - (height // 2)
    return f"{width}x{height}+{x}+{y}"


# ------------------------------------------------------------------------------



# pylint: disable=no-member

import argparse
import os
import re
import sys

from .rcparams import rcParams
from .utils import dhms_to_sec, die, is_dir_writable

poi_acc_cpu_default = 10

# ------------------------------------------------------------------------------


class Config:
    def __init__(self, app, version):
        self._app = app
        self.version = version

        self._filterfoos = []

        self.cores = 1
        self.toptype = None

        # Some versions of procps top have values that others don't
        # So far it's either mem_cached or mem_available
        self.mem_cached_or_available = "mem_cached"
        self.mem_unit = "KiB"

        self.has_cpu_column = False
        self.has_cpu_rows = False
        self.has_full_task_info = False
        self.has_swap = False

        self.poi_acc_cpu = None
        self.poi_categories = "categories: "  # subtitle of processes graph key
        self.profiling_tag = None
        self.start_time = None
        self.stop_time = None
        self.threads_not_tasks = False

        self._parse_commandline_arguments()
        self._handle_commandline_arguments()

    # --------------------------------------------------------------------------

    def _parse_commandline_arguments(self):
        self.parser = argparse.ArgumentParser(
            description=(
                "Munge data logged from the top utility in to graphs using"
                " GnuPlot.\nNote that Processes of Interest (POI) is a common-or-garden"
                " bucket, so however you select the processes, once they're in the"
                " bucket, they're just another process sloshing around in the bucket."
            )
        )

        self.parser.add_argument(
            "-f",
            "--file",
            dest="toplog_filename",
            metavar="PATH",
            default="top.log",
            help="Name of the file to munge (default: top.log)",
        )

        self.parser.add_argument(
            "-s",
            "--start",
            dest="start_time",
            metavar="TIMESTAMP",
            help=(
                "Start with timestamp TIMESTAMP ([D:]HH:MM:SS or >[[[D:]H+:]M+:]SS"
                " where the prefix '>' indicates an offset from the earliest log entry)"
            ),
        )

        self.parser.add_argument(
            "-S",
            "--stop",
            dest="stop_time",
            metavar="TIMESTAMP",
            help=(
                "Stop  with timestamp TIMESTAMP ([D:]HH:MM:SS or (+|<)[[[D:]H+:]M+:]SS"
                " where the prefix '<'indicates an offset from the earliest log entry"
                " and the + prefix an offset from the starting entry's timestamp)"
            ),
        )

        self.parser.add_argument(
            "-c",
            "--acc-cpu",
            dest="poi_acc_cpu",
            metavar="N",
            nargs="?",
            type=int,
            const=poi_acc_cpu_default,
            help="Top N processes ranked by accumulated CPU use (default: 10)",
        )

        self.parser.add_argument(
            "-m",
            "--acc-mem",
            dest="poi_acc_mem",
            metavar="N",
            nargs="?",
            type=int,
            const=10,
            help="Top N processes ranked by accumulated MEM use (default: 10)",
        )

        self.parser.add_argument(
            "--peak-cpu",
            dest="poi_peak_cpu",
            metavar="N",
            nargs="?",
            type=int,
            const=10,
            help="Top N processes ranked by peak CPU use (default: 10)",
        )

        self.parser.add_argument(
            "--peak-mem",
            dest="poi_peak_mem",
            metavar="N",
            nargs="?",
            type=int,
            const=10,
            help="Top N processes ranked by peak MEM use (default: 10)",
        )

        self.parser.add_argument(
            "--pct-cpu",
            dest="poi_cpu",
            metavar="PCT",
            nargs="?",
            type=int,
            const=20,
            help=(
                "Any process using more than PCT%% of memory will be graphed"
                " (default: 20)"
            ),
        )

        self.parser.add_argument(
            "--pct-mem",
            dest="poi_mem",
            metavar="PCT",
            nargs="?",
            type=int,
            const=3,
            help=(
                "Any process using more than PCT%% of cpu will be graphed (default: 3)"
            ),
        )

        self.parser.add_argument(
            "--prio",
            dest="poi_prio",
            metavar="cmpPRIO",
            nargs="?",
            type=str,
            const="=RT",
            help=(
                "Any process with priority =, <=, >=, <, or > to PRIO (default: '=RT',"
                " note the prefixed comparison operator)"
            ),
        )

        self.parser.add_argument(
            "-C",
            "--only-proc-cpu",
            dest="include_process_mem",
            action="store_false",
            help="Don't plot processes' mem info",
        )

        self.parser.add_argument(
            "-M",
            "--only-proc-mem",
            dest="include_process_cpu",
            action="store_false",
            help="Don't plot processes' cpu info",
        )

        self.parser.add_argument(
            "--with-cpu-steal",
            dest="with_cpu_steal",
            action="store_true",
            help="Plot CPU steal data",
        )

        self.parser.add_argument(
            "poi_regex",
            metavar="REGEX",
            nargs="?",
            help="Python style regex for names of processes to graph",
        )

        self.parser.add_argument(
            "-I",
            "--ignore",
            dest="ignore_processes_regex",
            metavar="REGEX",
            help="Python style regex for names of processes to ignore",
        )

        self.parser.add_argument(
            "-i",
            dest="ignore_case",
            action="store_true",
            help="Use case insensitive matching",
        )

        self.parser.add_argument(
            "--dont-plot-cpu-lines",
            dest="plot_poi_cpu_lines",
            action="store_false",
            help=(
                "Don't display the individual processes' cpu usage. Implies"
                " --plot-cpu-sum"
            ),
        )

        self.parser.add_argument(
            "--dont-plot-mem-lines",
            dest="plot_poi_mem_lines",
            action="store_false",
            help=(
                "Don't display the individual processes' mem usage. Implies"
                " --plot-mem-sum"
            ),
        )

        self.parser.add_argument(
            "--plot-cpu-sum",
            dest="plot_poi_cpu_sum",
            action="store_true",
            help="Add a line for the sum of the ploted processes' cpu usage",
        )

        self.parser.add_argument(
            "--plot-mem-sum",
            dest="plot_poi_mem_sum",
            action="store_true",
            help="Add a line for the sum of the ploted processes' mem usage",
        )

        self.parser.add_argument(
            "-G",
            "--no-graph",
            dest="do_graph",
            action="store_false",
            help=(
                'Don\'t plot a graph. Useful with "-v" to get just get info from stdout'
            ),
        )

        self.parser.add_argument(
            "-g",
            "--graph",
            dest="which_graphs",
            default="",
            metavar="STR",
            help=(
                "Display pane/s of the overview graph (1-4), the cpu data by core (C),"
                " or the poi data by core (c). E.g. '-g13' to display panes 1 and 3."
                " Also accepts 'q' and 'Q' for debugging purposes."
            ),
        )

        self.parser.add_argument(
            "--display-cpu-markers",
            action="store_true",
            help=(
                "Adding core markers to the lines on the CPU graph takes a long time,"
                " so they are not displayed by default. The overview figure never"
                " displays them. Use this option to have the separate CPU figure"
                " display them."
            ),
        )

        self.parser.add_argument(
            "--rcParams",
            dest="display_rc_params",
            action="store_true",
            help="Display MatPlotLib rcParams and exit",
        )

        self.parser.add_argument(
            "-O",
            "--outputdir",
            dest="output_dir",
            metavar="DIR",
            help="Location for saving files.",
        )

        self.parser.add_argument(
            "-l",
            "--list",
            dest="list_processes",
            action="count",
            default=0,
            help=(
                "List the processes recorded in the top logs (-ll and -lll increase"
                " info)"
            ),
        )

        self.parser.add_argument(
            "--nowrite",
            dest="allow_write",
            action="store_false",
            default=True,
            help="Don't write data. For debugging; use with --tmpdir DIR",
        )

        self.parser.add_argument(
            "--tmpdir",
            dest="tmpdir",
            metavar="DIR",
            help=(
                "Unmanaged directory to store temporary files in. For debugging or the"
                " curious."
            ),
        )

        try:
            # pylint: disable=import-outside-toplevel,unused-import
            from .profiler import Profiler  # noqa: F401 unused import
        except ImportError:
            pass
        else:
            self.parser.add_argument(
                "-P",
                "--profile",
                dest="profiling_tag",
                metavar="STR",
                nargs="?",
                default=None,
                const="",
                help=(
                    "Run with profiling turned on. If STR is given, this will be used"
                    " to tag in the filenames of emitted .prof files."
                ),
            )

        self.parser.add_argument(
            "-v",
            "--verbose",
            dest="verbosity",
            action="count",
            default=0,
            help="Increase verbosity to stdout (can be given multiple times)",
        )

        self.parser.add_argument(
            "-V", "--version", action="version", version=self.version
        )

        self.parser.parse_args(namespace=self)

    # --------------------------------------------------------------------------

    def filterfoo(self, command, pid, timestamp):
        for func in self._filterfoos:
            if func(command, pid, timestamp):
                return True
        return False

    # --------------------------------------------------------------------------

    def _handle_commandline_arguments(self):
        if self.display_rc_params:
            rcParams()
            sys.exit(0)

        # Preflight checks and conversions

        if not self.plot_poi_cpu_lines:
            self.plot_poi_cpu_sum = True
            if "c" in self.which_graphs or "C" in self.which_graphs:
                die("'-g [cC]' and '--dont-graph-cpu-lines' are incompatible.")

        if not self.plot_poi_mem_lines:
            self.plot_poi_mem_sum = True

        if not os.path.exists(self.toplog_filename):
            self.parser.print_usage()
            die(f"Can't see a file called '{self.toplog_filename}'.")

        if self.output_dir is not None:
            if not os.path.isdir(self.output_dir):
                if not os.path.exists(self.output_dir):
                    die(f"Can't see a directory called '{self.output_dir}'.")
                else:
                    die(f"'{self.output_dir}' is not a directory.")
            try:
                is_dir_writable(self.output_dir)
            except OSError as e:
                die(f"Can't write to '{self.output_dir}' : {e}.")

        if not self.include_process_cpu and not self.include_process_mem:
            die("Can't disable processing both processes' cpu and processes' mem")

        if self.poi_prio:
            if self.poi_prio[0:1] not in "=<>":
                die("--prio argument must start with =, <=, >=, <, or >")

        if self.start_time:
            check = re.compile(r"^([>0-9])")
            result = check.match(self.start_time[0:1])
            if not result:
                die(
                    "-s | --start : offset requires leading '>'. Dunno about"
                    f" {self.start_time[0:1]}"
                )

            if result.group(0) != ">":
                self.start_time = dhms_to_sec(self.start_time)

        if self.stop_time:
            check = re.compile(r"^([<+0-9])")
            result = check.match(self.stop_time[0:1])
            if not result:
                die(
                    "-S | --stop : offset requires leading '+' or '<'. Dunno about"
                    f" {self.stop_time[0:1]}"
                )

            if result.group(0) != "+" and result.group(0) != "<":
                self.stop_time = dhms_to_sec(self.stop_time)

        # ----------------------------------------------------------------------
        # Processes of interest are identified by filtering through a list of
        # lambda functions and by creating top-N lists of cpu and mem values

        no_specific_interest = not (
            self.poi_cpu
            or self.poi_mem
            or self.poi_acc_cpu
            or self.poi_acc_mem
            or self.poi_peak_cpu
            or self.poi_peak_mem
            or self.poi_prio
            or self.poi_regex
        )

        if self.poi_mem:
            self._filterfoos.append(
                lambda command, pid, timestamp: float(
                    self._app.processes[command][pid]["timestamps"][timestamp]["mem"]
                )
                > self.poi_mem
            )
            self.poi_categories += f"mem>{self.poi_mem} "

        if self.poi_cpu:
            self._filterfoos.append(
                lambda command, pid, timestamp: float(
                    self._app.processes[command][pid]["timestamps"][timestamp]["cpu"]
                )
                > self.poi_cpu
            )
            self.poi_categories += f"cpu>{self.poi_cpu} "

        if self.poi_prio:
            if self.poi_prio[0:1] == "=" or self.poi_prio[0:2] == "==":
                i = 2 if self.poi_prio[0:2] == "==" else 1
                self._filterfoos.append(
                    lambda command, pid, timestamp: self._app.processes[command][pid][
                        "timestamps"
                    ][timestamp]["priority"]
                    == self.poi_prio[i:]
                )
            elif self.poi_prio[0:2] == "<=":
                self._filterfoos.append(
                    lambda command, pid, timestamp: self._app.processes[command][pid][
                        "timestamps"
                    ][timestamp]["priority"]
                    <= self.poi_prio[2:]
                )
            elif self.poi_prio[0:2] == ">=":
                self._filterfoos.append(
                    lambda command, pid, timestamp: self._app.processes[command][pid][
                        "timestamps"
                    ][timestamp]["priority"]
                    >= self.poi_prio[2:]
                )
            elif self.poi_prio[0:1] == "<":
                self._filterfoos.append(
                    lambda command, pid, timestamp: self._app.processes[command][pid][
                        "timestamps"
                    ][timestamp]["priority"]
                    < self.poi_prio[1:]
                )
            elif self.poi_prio[0:1] == ">":
                self._filterfoos.append(
                    lambda command, pid, timestamp: self._app.processes[command][pid][
                        "timestamps"
                    ][timestamp]["priority"]
                    > self.poi_prio[1:]
                )

            self.poi_categories += f"prio{self.poi_prio} "

        if self.poi_regex:
            flags = 0
            if self.ignore_case:
                flags = re.IGNORECASE

            self.poi_regex = re.compile(self.poi_regex, flags)
            self._filterfoos.append(
                lambda command, pid, timestamp: self.poi_regex.match(command) is not None
            )
            self.poi_categories += (
                f"/{self.poi_regex}/{'i' if self.ignore_case else ''} "
            )

        if self.poi_acc_cpu or no_specific_interest:
            if not self.poi_acc_cpu:
                self.poi_acc_cpu = poi_acc_cpu_default

            self.poi_categories += f"cpu+{self.poi_acc_cpu} "

        if self.poi_acc_mem:
            self.poi_categories += f"mem+{self.poi_acc_mem} "

        if self.poi_peak_cpu:
            self.poi_categories += f"cpu^{self.poi_peak_cpu} "

        if self.poi_peak_mem:
            self.poi_categories += f"mem^{self.poi_peak_mem} "

        self.poi_categories = self.poi_categories.strip()


# ------------------------------------------------------------------------------
